import torch
import os

import ast
import json
import os.path as osp
import re
import shutil
import sys
import tempfile
from copy import deepcopy
from importlib import import_module

import yaml

from ...video_utils import get_video_details, load_frames_from_video
import numpy as np

from torchvision import transforms
from torchvision.transforms import InterpolationMode

from .models.umt import UMT
from .shared_utils import setup_model

from easydict import EasyDict

from .models.criterions import get_sim

BASE_KEY = "_base_"
BASE_CONFIG = {}


class Config(object):
    """config"""

    @classmethod
    def pretty_text(cls, cfg: dict, indent=2) -> str:
        """format dict to a string

        Args:
            cfg (EasyDict): the params.

        Returns: The string to display.

        """
        msg = "{\n"
        for i, (k, v) in enumerate(cfg.items()):
            if isinstance(v, dict):
                v = cls.pretty_text(v, indent + 4)
            spaces = " " * indent
            msg += spaces + "{}: {}".format(k, v)
            if i == len(cfg) - 1:
                msg += " }"
            else:
                msg += "\n"
        return msg

    @classmethod
    def dump(cls, cfg, savepath=None):
        """dump cfg to `json` file.

        Args:
            cfg (dict): The dict to dump.
            savepath (str): The filepath to save the dumped dict.

        Returns: TODO

        """
        if savepath is None:
            savepath = osp.join(cfg.WORKSPACE, "config.json")
        json.dump(cfg, open(savepath, "w"), indent=2)

    @classmethod
    def get_config(
        cls, model_name: str, model_path: str, pretrained_path: str
    ) -> EasyDict:
        cfg = EasyDict(BASE_CONFIG)
        current_dir = os.path.dirname(os.path.abspath(__file__))  # Get the current script's directory
        target_file = os.path.join(current_dir, f"{model_name}.py")  # Construct the new file path
        cfg_from_file = cls.from_file(target_file)
        cfg = merge_a_into_b(cfg_from_file, cfg)
        cfg = cls.merge_list(cfg, ["model.vision_encoder.pretrained", pretrained_path])
        cfg = cls.merge_list(cfg, ["pretrained_path", model_path])
        cfg = eval_dict_leaf(cfg)

        # update some keys to make them show at the last
        for k in BASE_CONFIG:
            cfg[k] = cfg.pop(k)
        return cfg

    @classmethod
    def from_file(cls, filepath: str) -> EasyDict:
        """Build config from file. Supported filetypes: `.py`,`.yaml`,`.json`.

        Args:
            filepath (str): The config file path.

        Returns: TODO

        """
        filepath = osp.abspath(osp.expanduser(filepath))
        if not osp.isfile(filepath):
            raise IOError(f"File does not exist: {filepath}")
        if filepath.endswith(".py"):
            with tempfile.TemporaryDirectory() as temp_config_dir:

                shutil.copytree(
                    osp.dirname(filepath), osp.join(temp_config_dir, "tmp_config")
                )
                sys.path.insert(0, temp_config_dir)
                mod = import_module(
                    "tmp_config." + osp.splitext(osp.basename(filepath))[0]
                )
                # mod = import_module(temp_module_name)
                sys.path.pop(0)
                cfg_dict = {
                    name: value
                    for name, value in mod.__dict__.items()
                    if not name.startswith("__")
                }
                for k in list(sys.modules.keys()):
                    if "tmp_config" in k:
                        del sys.modules[k]
        elif filepath.endswith((".yml", ".yaml")):
            cfg_dict = yaml.load(open(filepath, "r"), Loader=yaml.Loader)
        elif filepath.endswith(".json"):
            cfg_dict = json.load(open(filepath, "r"))
        else:
            raise IOError("Only py/yml/yaml/json type are supported now!")

        cfg_text = filepath + "\n"
        with open(filepath, "r") as f:
            cfg_text += f.read()

        if BASE_KEY in cfg_dict:  # load configs in `BASE_KEY`
            cfg_dir = osp.dirname(filepath)
            base_filename = cfg_dict.pop(BASE_KEY)
            base_filename = (
                base_filename if isinstance(base_filename, list) else [base_filename]
            )

            cfg_dict_list = list()
            for f in base_filename:
                _cfg_dict = Config.from_file(osp.join(cfg_dir, f))
                cfg_dict_list.append(_cfg_dict)

            base_cfg_dict = dict()
            for c in cfg_dict_list:
                if len(base_cfg_dict.keys() & c.keys()) > 0:
                    raise KeyError("Duplicate key is not allowed among bases")
                base_cfg_dict.update(c)

            cfg_dict = merge_a_into_b(cfg_dict, base_cfg_dict)

        return EasyDict(cfg_dict)

    @classmethod
    def merge_list(cls, cfg, opts: list):
        """merge commandline opts.

        Args:
            cfg: (dict): The config to be merged.
            opts (list): The list to merge. Format: [key1, name1, key2, name2,...].
                The keys can be nested. For example, ["a.b", v] will be considered
                as `dict(a=dict(b=v))`.

        Returns: dict.

        """
        assert len(opts) % 2 == 0, f"length of opts must be even. Got: {opts}"
        for i in range(0, len(opts), 2):
            full_k, v = opts[i], opts[i + 1]
            keys = full_k.split(".")
            sub_d = cfg
            for i, k in enumerate(keys):
                if not hasattr(sub_d, k):
                    raise ValueError(
                        f"The key {k} not exist in the config. Full key:{full_k}"
                    )
                if i != len(keys) - 1:
                    sub_d = sub_d[k]
                else:
                    sub_d[k] = v
        return cfg


def merge_a_into_b(a, b, inplace=False):
    """The values in a will override values in b.

    Args:
        a (dict): source dict.
        b (dict): target dict.

    Returns: dict. recursively merge dict a into dict b.

    """
    if not inplace:
        b = deepcopy(b)
    for key in a:
        if key in b:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                b[key] = merge_a_into_b(a[key], b[key], inplace=True)
            else:
                b[key] = a[key]
        else:
            b[key] = a[key]
    return b


def eval_dict_leaf(d, orig_dict=None):
    """eval values of dict leaf.

    Args:
        d (dict): The dict to eval.

    Returns: dict.

    """
    if orig_dict is None:
        orig_dict = d
    for k, v in d.items():
        if not isinstance(v, dict):
            d[k] = eval_string(v, orig_dict)
        else:
            eval_dict_leaf(v, orig_dict)
    return d


def eval_string(string, d):
    """automatically evaluate string to corresponding types.

    For example:
        not a string  -> return the original input
        '0'  -> 0
        '0.2' -> 0.2
        '[0, 1, 2]' -> [0,1,2]
        'eval(1+2)' -> 3
        'eval(range(5))' -> [0,1,2,3,4]
        '${a}' -> d.a



    Args:
        string (str): The value to evaluate.
        d (dict): The

    Returns: the corresponding type

    """
    if not isinstance(string, str):
        return string
    # if len(string) > 1 and string[0] == "[" and string[-1] == "]":
    #     return eval(string)
    if string[0:5] == "eval(":
        return eval(string[5:-1])

    s0 = string
    s1 = re.sub(r"\${(.*)}", r"d.\1", s0)
    if s1 != s0:
        while s1 != s0:
            s0 = s1
            s1 = re.sub(r"\${(.*)}", r"d.\1", s0)
        return eval(s1)

    try:
        v = ast.literal_eval(string)
    except:
        v = string
    return v


def extract_text_feats(texts, max_txt_l, tokenizer, model, device):
    text_feats = []
    text_atts = []

    text = texts
    text_input = tokenizer(
        text,
        padding="max_length",
        truncation=True,
        max_length=max_txt_l,
        return_tensors="pt",
    ).to(device)

    text_feat = model.encode_text(text_input)[0]
    text_feats.append(text_feat)
    text_atts.append(text_input.attention_mask)

    text_feats = torch.cat(text_feats, dim=0)
    text_atts = torch.cat(text_atts, dim=0)
    return text_feats, text_atts


def extract_vision_feats(image_paths, transforms, model, device, num_frames=4):
    image_feats_all = []
    pooled_image_feats_all = []
    image = []
    for data_path in image_paths:
        total_frames, original_fps, video_duration = get_video_details(data_path)
        uniform_sampling = min(num_frames, total_frames)
        all_indices = np.linspace(0, total_frames - 1, uniform_sampling, dtype=int)
        frames = load_frames_from_video(data_path, all_indices, "decord", True)
        frames = frames.permute(0, 3, 1, 2)  # (T, C, H, W), torch.uint8
        frames = transforms(frames)
        image.append(frames)
    image = torch.stack(image, dim=0)
    image = image.to(device, non_blocking=True)
    image_feat, pooled_image_feat = model.encode_vision(image, test=True)
    image_feat = image_feat.unsqueeze(1)  # (bsz, 1, #frm*L, d)
    image_feats_all.append(image_feat)
    pooled_image_feats_all.append(pooled_image_feat)
    image_feats_all = torch.cat(image_feats_all, dim=0)

    pooled_image_feats_all = torch.cat(pooled_image_feats_all, dim=0)
    return image_feats_all, pooled_image_feats_all


def create_transforms(image_res=224):
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)
    normalize = transforms.Normalize(mean, std)
    # loaded images and videos are torch.Tensor of torch.uint8 format,
    # ordered as (T, 1 or 3, H, W) where T=1 for image
    type_transform = transforms.Lambda(lambda x: x.float().div(255.0))

    test_transform = transforms.Compose(
        [
            transforms.Resize(
                (image_res, image_res),
                interpolation=InterpolationMode.BICUBIC,
            ),
            type_transform,
            normalize,
        ]
    )
    return test_transform


@torch.no_grad()
def evaluation(
    texts,
    image_paths,
    transforms,
    model,
    tokenizer,
    device,
    num_frames=4,
    max_txt_l=32,
):
    text_feats, text_atts = extract_text_feats(
        texts, max_txt_l, tokenizer, model, device
    )  # (bsz, Lt, d), (bsz, Lt)

    image_feats, pooled_image_feats = extract_vision_feats(
        image_paths, transforms, model, device, num_frames=num_frames
    )  # (bsz, 1, #frm*Li, d) or (bsz, #frm, Li, d), (bsz, #frm, d)
    _pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)
    i2t_scores, _ = get_sim(
        model.vision_proj(_pooled_image_feats), model.text_proj(text_feats[:, 0])
    )

    text_encoder = model.get_text_encoder()
    encoder_output = image_feats[:, 0].to(device, non_blocking=True)
    encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(device, non_blocking=True)

    output = text_encoder(
        encoder_embeds=text_feats,
        attention_mask=text_atts,
        encoder_hidden_states=encoder_output,
        encoder_attention_mask=encoder_att,
        return_dict=True,
        mode="fusion",
    )
    itm_outputs = output.last_hidden_state[:, 0]
    itm_embeds = torch.cat([itm_outputs], dim=0)
    itm_scores = model.itm_head(itm_embeds)[:, 1]

    return (
        itm_scores,
        i2t_scores.diagonal(),
    )


def download_umt(model_name, pretrained_ckpt_name, cache_dir, device="cuda"):
    repo_id = f"zhiqiulin/{model_name}"
    filename = f"{model_name}.pth"
    model_path = os.path.join(cache_dir, model_name) + ".pth"
    if not os.path.exists(model_path):
        hf_url = f"https://huggingface.co/{repo_id}/resolve/main/{filename}"
        os.system(f"wget -O {model_path} {hf_url}")

    pretrained_repo_id = f"zhiqiulin/{pretrained_ckpt_name}"
    pretrained_filename = f"{pretrained_ckpt_name}.pth"
    pretrained_ckpt_path = os.path.join(cache_dir, f"{pretrained_ckpt_name}.pth")
    if not os.path.exists(pretrained_ckpt_path):
        hf_url = f"https://huggingface.co/{pretrained_repo_id}/resolve/main/{pretrained_filename}"
        os.system(f"wget -O {pretrained_ckpt_path} {hf_url}")

    config = Config.get_config(
        model_name=model_name,
        model_path=model_path,
        pretrained_path=pretrained_ckpt_path,
    )
    assert config.evaluate

    (
        model,
        tokenizer,
    ) = setup_model(
        config,
        model_cls=UMT,
        has_decoder=False,
        pretrain=False,
        # find_unused_parameters=False,
        device=device
    )
    model = model.eval()
    tokenizer = tokenizer
    config = config
    return (
        model_path,
        pretrained_ckpt_path,
        model,
        tokenizer,
        config,
        create_transforms(224),
    )



"""
 * Copyright (c) 2023, salesforce.com, inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see LICENSE.txt file in the repo root or https://opensource.org/licenses/BSD-3-Clause
 * By Junnan Li
 * Based on huggingface code base
 * https://github.com/huggingface/transformers/blob/v4.15.0/src/transformers/models/bert
"""

import math
import os
import warnings
from dataclasses import dataclass
from typing import Optional, Tuple, Dict, Any

import torch
from torch import Tensor, device, dtype, nn
import torch.utils.checkpoint
from torch import nn
from torch.nn import CrossEntropyLoss
import torch.nn.functional as F

from transformers.activations import ACT2FN
from transformers.file_utils import (
    ModelOutput,
)
from transformers.modeling_outputs import (
    BaseModelOutputWithPastAndCrossAttentions,
    BaseModelOutputWithPoolingAndCrossAttentions,
    CausalLMOutputWithCrossAttentions,
    MaskedLMOutput,
    MultipleChoiceModelOutput,
    NextSentencePredictorOutput,
    QuestionAnsweringModelOutput,
    SequenceClassifierOutput,
    TokenClassifierOutput,
)
from transformers.modeling_utils import (
    PreTrainedModel,
    apply_chunking_to_forward,
    find_pruneable_heads_and_indices,
    prune_linear_layer,
)
from transformers.utils import logging
from transformers.models.bert.configuration_bert import BertConfig

logger = logging.get_logger(__name__)


class BertEmbeddings(nn.Module):
    """Construct the embeddings from word and position embeddings."""

    def __init__(self, config):
        super().__init__()
        self.word_embeddings = nn.Embedding(
            config.vocab_size, config.hidden_size, padding_idx=config.pad_token_id
        )
        self.position_embeddings = nn.Embedding(
            config.max_position_embeddings, config.hidden_size
        )

        # self.LayerNorm is not snake-cased to stick with TensorFlow model variable name and be able to load
        # any TensorFlow checkpoint file
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

        # position_ids (1, len position emb) is contiguous in memory and exported when serialized
        self.register_buffer(
            "position_ids", torch.arange(config.max_position_embeddings).expand((1, -1))
        )
        self.position_embedding_type = getattr(
            config, "position_embedding_type", "absolute"
        )

        self.config = config

    def forward(
        self,
        input_ids=None,
        position_ids=None,
        query_embeds=None,
        past_key_values_length=0,
    ):
        if input_ids is not None:
            seq_length = input_ids.size()[1]
        else:
            seq_length = 0

        if position_ids is None:
            position_ids = self.position_ids[
                :, past_key_values_length : seq_length + past_key_values_length
            ].clone()

        if input_ids is not None:
            embeddings = self.word_embeddings(input_ids)
            if self.position_embedding_type == "absolute":
                position_embeddings = self.position_embeddings(position_ids)
                embeddings = embeddings + position_embeddings

            if query_embeds is not None:
                embeddings = torch.cat((query_embeds, embeddings), dim=1)
        else:
            embeddings = query_embeds

        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)
        return embeddings


class BertSelfAttention(nn.Module):
    def __init__(self, config, is_cross_attention):
        super().__init__()
        self.config = config
        if config.hidden_size % config.num_attention_heads != 0 and not hasattr(
            config, "embedding_size"
        ):
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (config.hidden_size, config.num_attention_heads)
            )

        self.num_attention_heads = config.num_attention_heads
        self.attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        self.query = nn.Linear(config.hidden_size, self.all_head_size)
        if is_cross_attention:
            self.key = nn.Linear(config.encoder_width, self.all_head_size)
            self.value = nn.Linear(config.encoder_width, self.all_head_size)
        else:
            self.key = nn.Linear(config.hidden_size, self.all_head_size)
            self.value = nn.Linear(config.hidden_size, self.all_head_size)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        self.position_embedding_type = getattr(
            config, "position_embedding_type", "absolute"
        )
        if (
            self.position_embedding_type == "relative_key"
            or self.position_embedding_type == "relative_key_query"
        ):
            self.max_position_embeddings = config.max_position_embeddings
            self.distance_embedding = nn.Embedding(
                2 * config.max_position_embeddings - 1, self.attention_head_size
            )
        self.save_attention = False

    def save_attn_gradients(self, attn_gradients):
        self.attn_gradients = attn_gradients

    def get_attn_gradients(self):
        return self.attn_gradients

    def save_attention_map(self, attention_map):
        self.attention_map = attention_map

    def get_attention_map(self):
        return self.attention_map

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
    ):

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        is_cross_attention = encoder_hidden_states is not None

        if is_cross_attention:
            key_layer = self.transpose_for_scores(self.key(encoder_hidden_states))
            value_layer = self.transpose_for_scores(self.value(encoder_hidden_states))
            attention_mask = encoder_attention_mask
        elif past_key_value is not None:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))
            key_layer = torch.cat([past_key_value[0], key_layer], dim=2)
            value_layer = torch.cat([past_key_value[1], value_layer], dim=2)
        else:
            key_layer = self.transpose_for_scores(self.key(hidden_states))
            value_layer = self.transpose_for_scores(self.value(hidden_states))

        mixed_query_layer = self.query(hidden_states)

        query_layer = self.transpose_for_scores(mixed_query_layer)

        past_key_value = (key_layer, value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        if (
            self.position_embedding_type == "relative_key"
            or self.position_embedding_type == "relative_key_query"
        ):
            seq_length = hidden_states.size()[1]
            position_ids_l = torch.arange(
                seq_length, dtype=torch.long, device=hidden_states.device
            ).view(-1, 1)
            position_ids_r = torch.arange(
                seq_length, dtype=torch.long, device=hidden_states.device
            ).view(1, -1)
            distance = position_ids_l - position_ids_r
            positional_embedding = self.distance_embedding(
                distance + self.max_position_embeddings - 1
            )
            positional_embedding = positional_embedding.to(
                dtype=query_layer.dtype
            )  # fp16 compatibility

            if self.position_embedding_type == "relative_key":
                relative_position_scores = torch.einsum(
                    "bhld,lrd->bhlr", query_layer, positional_embedding
                )
                attention_scores = attention_scores + relative_position_scores
            elif self.position_embedding_type == "relative_key_query":
                relative_position_scores_query = torch.einsum(
                    "bhld,lrd->bhlr", query_layer, positional_embedding
                )
                relative_position_scores_key = torch.einsum(
                    "bhrd,lrd->bhlr", key_layer, positional_embedding
                )
                attention_scores = (
                    attention_scores
                    + relative_position_scores_query
                    + relative_position_scores_key
                )

        attention_scores = attention_scores / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
            attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-1)(attention_scores)

        if is_cross_attention and self.save_attention:
            self.save_attention_map(attention_probs)
            attention_probs.register_hook(self.save_attn_gradients)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        attention_probs_dropped = self.dropout(attention_probs)

        # Mask heads if we want to
        if head_mask is not None:
            attention_probs_dropped = attention_probs_dropped * head_mask

        context_layer = torch.matmul(attention_probs_dropped, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)

        outputs = (
            (context_layer, attention_probs) if output_attentions else (context_layer,)
        )

        outputs = outputs + (past_key_value,)
        return outputs


class BertSelfOutput(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertAttention(nn.Module):
    def __init__(self, config, is_cross_attention=False):
        super().__init__()
        self.self = BertSelfAttention(config, is_cross_attention)
        self.output = BertSelfOutput(config)
        self.pruned_heads = set()

    def prune_heads(self, heads):
        if len(heads) == 0:
            return
        heads, index = find_pruneable_heads_and_indices(
            heads,
            self.self.num_attention_heads,
            self.self.attention_head_size,
            self.pruned_heads,
        )

        # Prune linear layers
        self.self.query = prune_linear_layer(self.self.query, index)
        self.self.key = prune_linear_layer(self.self.key, index)
        self.self.value = prune_linear_layer(self.self.value, index)
        self.output.dense = prune_linear_layer(self.output.dense, index, dim=1)

        # Update hyper params and store pruned heads
        self.self.num_attention_heads = self.self.num_attention_heads - len(heads)
        self.self.all_head_size = (
            self.self.attention_head_size * self.self.num_attention_heads
        )
        self.pruned_heads = self.pruned_heads.union(heads)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
    ):
        self_outputs = self.self(
            hidden_states,
            attention_mask,
            head_mask,
            encoder_hidden_states,
            encoder_attention_mask,
            past_key_value,
            output_attentions,
        )
        attention_output = self.output(self_outputs[0], hidden_states)

        outputs = (attention_output,) + self_outputs[
            1:
        ]  # add attentions if we output them
        return outputs


class BertIntermediate(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.intermediate_size)
        if isinstance(config.hidden_act, str):
            self.intermediate_act_fn = ACT2FN[config.hidden_act]
        else:
            self.intermediate_act_fn = config.hidden_act

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.intermediate_act_fn(hidden_states)
        return hidden_states


class BertOutput(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states, input_tensor):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class BertLayer(nn.Module):
    def __init__(self, config, layer_num):
        super().__init__()
        self.config = config
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = BertAttention(config)
        self.layer_num = layer_num
        if (
            self.config.add_cross_attention
            and layer_num % self.config.cross_attention_freq == 0
        ):
            self.crossattention = BertAttention(
                config, is_cross_attention=self.config.add_cross_attention
            )
            self.has_cross_attention = True
        else:
            self.has_cross_attention = False
        self.intermediate = BertIntermediate(config)
        self.output = BertOutput(config)

        self.intermediate_query = BertIntermediate(config)
        self.output_query = BertOutput(config)

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_value=None,
        output_attentions=False,
        query_length=0,
    ):
        # decoder uni-directional self-attention cached key/values tuple is at positions 1,2
        self_attn_past_key_value = (
            past_key_value[:2] if past_key_value is not None else None
        )
        self_attention_outputs = self.attention(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions=output_attentions,
            past_key_value=self_attn_past_key_value,
        )
        attention_output = self_attention_outputs[0]
        outputs = self_attention_outputs[1:-1]

        present_key_value = self_attention_outputs[-1]

        if query_length > 0:
            query_attention_output = attention_output[:, :query_length, :]

            if self.has_cross_attention:
                assert (
                    encoder_hidden_states is not None
                ), "encoder_hidden_states must be given for cross-attention layers"
                cross_attention_outputs = self.crossattention(
                    query_attention_output,
                    attention_mask,
                    head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    output_attentions=output_attentions,
                )
                query_attention_output = cross_attention_outputs[0]
                outputs = (
                    outputs + cross_attention_outputs[1:-1]
                )  # add cross attentions if we output attention weights

            layer_output = apply_chunking_to_forward(
                self.feed_forward_chunk_query,
                self.chunk_size_feed_forward,
                self.seq_len_dim,
                query_attention_output,
            )
            if attention_output.shape[1] > query_length:
                layer_output_text = apply_chunking_to_forward(
                    self.feed_forward_chunk,
                    self.chunk_size_feed_forward,
                    self.seq_len_dim,
                    attention_output[:, query_length:, :],
                )
                layer_output = torch.cat([layer_output, layer_output_text], dim=1)
        else:
            layer_output = apply_chunking_to_forward(
                self.feed_forward_chunk,
                self.chunk_size_feed_forward,
                self.seq_len_dim,
                attention_output,
            )
        outputs = (layer_output,) + outputs

        outputs = outputs + (present_key_value,)

        return outputs

    def feed_forward_chunk(self, attention_output):
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output

    def feed_forward_chunk_query(self, attention_output):
        intermediate_output = self.intermediate_query(attention_output)
        layer_output = self.output_query(intermediate_output, attention_output)
        return layer_output


class BertEncoder(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList(
            [BertLayer(config, i) for i in range(config.num_hidden_layers)]
        )

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=False,
        output_hidden_states=False,
        return_dict=True,
        query_length=0,
    ):
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = () if output_attentions else None
        all_cross_attentions = (
            () if output_attentions and self.config.add_cross_attention else None
        )

        next_decoder_cache = () if use_cache else None

        for i in range(self.config.num_hidden_layers):
            layer_module = self.layer[i]
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None
            past_key_value = past_key_values[i] if past_key_values is not None else None

            if getattr(self.config, "gradient_checkpointing", False) and self.training:

                if use_cache:
                    logger.warn(
                        "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(
                            *inputs, past_key_value, output_attentions, query_length
                        )

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    past_key_value,
                    output_attentions,
                    query_length,
                )

            hidden_states = layer_outputs[0]
            if use_cache:
                next_decoder_cache += (layer_outputs[-1],)
            if output_attentions:
                all_self_attentions = all_self_attentions + (layer_outputs[1],)
                all_cross_attentions = all_cross_attentions + (layer_outputs[2],)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    next_decoder_cache,
                    all_hidden_states,
                    all_self_attentions,
                    all_cross_attentions,
                ]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_decoder_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
            cross_attentions=all_cross_attentions,
        )


class BertPooler(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        self.activation = nn.Tanh()

    def forward(self, hidden_states):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = hidden_states[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class BertPredictionHeadTransform(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.hidden_size, config.hidden_size)
        if isinstance(config.hidden_act, str):
            self.transform_act_fn = ACT2FN[config.hidden_act]
        else:
            self.transform_act_fn = config.hidden_act
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)

    def forward(self, hidden_states):
        hidden_states = self.dense(hidden_states)
        hidden_states = self.transform_act_fn(hidden_states)
        hidden_states = self.LayerNorm(hidden_states)
        return hidden_states


class BertLMPredictionHead(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.transform = BertPredictionHeadTransform(config)

        # The output weights are the same as the input embeddings, but there is
        # an output-only bias for each token.
        self.decoder = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        self.bias = nn.Parameter(torch.zeros(config.vocab_size))

        # Need a link between the two variables so that the bias is correctly resized with `resize_token_embeddings`
        self.decoder.bias = self.bias

    def forward(self, hidden_states):
        hidden_states = self.transform(hidden_states)
        hidden_states = self.decoder(hidden_states)
        return hidden_states


class BertOnlyMLMHead(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.predictions = BertLMPredictionHead(config)

    def forward(self, sequence_output):
        prediction_scores = self.predictions(sequence_output)
        return prediction_scores


class BertPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = BertConfig
    base_model_prefix = "bert"
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def _init_weights(self, module):
        """Initialize the weights"""
        if isinstance(module, (nn.Linear, nn.Embedding)):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)
        if isinstance(module, nn.Linear) and module.bias is not None:
            module.bias.data.zero_()


class BertModel(BertPreTrainedModel):
    """
    The model can behave as an encoder (with only self-attention) as well as a decoder, in which case a layer of
    cross-attention is added between the self-attention layers, following the architecture described in `Attention is
    all you need <https://arxiv.org/abs/1706.03762>`__ by Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit,
    Llion Jones, Aidan N. Gomez, Lukasz Kaiser and Illia Polosukhin.
    argument and :obj:`add_cross_attention` set to :obj:`True`; an :obj:`encoder_hidden_states` is then expected as an
    input to the forward pass.
    """

    def __init__(self, config, add_pooling_layer=False):
        super().__init__(config)
        self.config = config

        self.embeddings = BertEmbeddings(config)

        self.encoder = BertEncoder(config)

        self.pooler = BertPooler(config) if add_pooling_layer else None

        self.init_weights()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def get_extended_attention_mask(
        self,
        attention_mask: Tensor,
        input_shape: Tuple[int],
        device: device,
        is_decoder: bool,
        has_query: bool = False,
    ) -> Tensor:
        """
        Makes broadcastable attention and causal masks so that future and masked tokens are ignored.

        Arguments:
            attention_mask (:obj:`torch.Tensor`):
                Mask with ones indicating tokens to attend to, zeros for tokens to ignore.
            input_shape (:obj:`Tuple[int]`):
                The shape of the input to the model.
            device: (:obj:`torch.device`):
                The device of the input to the model.

        Returns:
            :obj:`torch.Tensor` The extended attention mask, with a the same dtype as :obj:`attention_mask.dtype`.
        """
        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        if attention_mask.dim() == 3:
            extended_attention_mask = attention_mask[:, None, :, :]
        elif attention_mask.dim() == 2:
            # Provided a padding mask of dimensions [batch_size, seq_length]
            # - if the model is a decoder, apply a causal mask in addition to the padding mask
            # - if the model is an encoder, make the mask broadcastable to [batch_size, num_heads, seq_length, seq_length]
            if is_decoder:
                batch_size, seq_length = input_shape

                seq_ids = torch.arange(seq_length, device=device)
                causal_mask = (
                    seq_ids[None, None, :].repeat(batch_size, seq_length, 1)
                    <= seq_ids[None, :, None]
                )

                # add a prefix ones mask to the causal mask
                # causal and attention masks must have same type with pytorch version < 1.3
                causal_mask = causal_mask.to(attention_mask.dtype)

                if causal_mask.shape[1] < attention_mask.shape[1]:
                    prefix_seq_len = attention_mask.shape[1] - causal_mask.shape[1]
                    if has_query:  # UniLM style attention mask
                        causal_mask = torch.cat(
                            [
                                torch.zeros(
                                    (batch_size, prefix_seq_len, seq_length),
                                    device=device,
                                    dtype=causal_mask.dtype,
                                ),
                                causal_mask,
                            ],
                            axis=1,
                        )
                    causal_mask = torch.cat(
                        [
                            torch.ones(
                                (batch_size, causal_mask.shape[1], prefix_seq_len),
                                device=device,
                                dtype=causal_mask.dtype,
                            ),
                            causal_mask,
                        ],
                        axis=-1,
                    )
                extended_attention_mask = (
                    causal_mask[:, None, :, :] * attention_mask[:, None, None, :]
                )
            else:
                extended_attention_mask = attention_mask[:, None, None, :]
        else:
            raise ValueError(
                "Wrong shape for input_ids (shape {}) or attention_mask (shape {})".format(
                    input_shape, attention_mask.shape
                )
            )

        # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
        # masked positions, this operation will create a tensor which is 0.0 for
        # positions we want to attend and -10000.0 for masked positions.
        # Since we are adding it to the raw scores before the softmax, this is
        # effectively the same as removing these entirely.
        extended_attention_mask = extended_attention_mask.to(
            dtype=self.dtype
        )  # fp16 compatibility
        extended_attention_mask = (1.0 - extended_attention_mask) * -10000.0
        return extended_attention_mask

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        position_ids=None,
        head_mask=None,
        query_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        is_decoder=False,
    ):
        r"""
        encoder_hidden_states  (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in ``[0, 1]``:
            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))` of length :obj:`config.n_layers` with each tuple having 4 tensors of shape :obj:`(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.
            If :obj:`past_key_values` are used, the user can optionally input only the last :obj:`decoder_input_ids`
            (those that don't have their past key value states given to this model) of shape :obj:`(batch_size, 1)`
            instead of all :obj:`decoder_input_ids` of shape :obj:`(batch_size, sequence_length)`.
        use_cache (:obj:`bool`, `optional`):
            If set to :obj:`True`, :obj:`past_key_values` key value states are returned and can be used to speed up
            decoding (see :obj:`past_key_values`).
        """
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        # use_cache = use_cache if use_cache is not None else self.config.use_cache

        if input_ids is None:
            assert (
                query_embeds is not None
            ), "You have to specify query_embeds when input_ids is None"

        # past_key_values_length
        past_key_values_length = (
            past_key_values[0][0].shape[2] - self.config.query_length
            if past_key_values is not None
            else 0
        )

        query_length = query_embeds.shape[1] if query_embeds is not None else 0

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            query_embeds=query_embeds,
            past_key_values_length=past_key_values_length,
        )

        input_shape = embedding_output.size()[:-1]
        batch_size, seq_length = input_shape
        device = embedding_output.device

        if attention_mask is None:
            attention_mask = torch.ones(
                ((batch_size, seq_length + past_key_values_length)), device=device
            )

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        if is_decoder:
            extended_attention_mask = self.get_extended_attention_mask(
                attention_mask,
                input_ids.shape,
                device,
                is_decoder,
                has_query=(query_embeds is not None),
            )
        else:
            extended_attention_mask = self.get_extended_attention_mask(
                attention_mask, input_shape, device, is_decoder
            )

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if encoder_hidden_states is not None:
            if type(encoder_hidden_states) == list:
                encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states[
                    0
                ].size()
            else:
                (
                    encoder_batch_size,
                    encoder_sequence_length,
                    _,
                ) = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)

            if type(encoder_attention_mask) == list:
                encoder_extended_attention_mask = [
                    self.invert_attention_mask(mask) for mask in encoder_attention_mask
                ]
            elif encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
                encoder_extended_attention_mask = self.invert_attention_mask(
                    encoder_attention_mask
                )
            else:
                encoder_extended_attention_mask = self.invert_attention_mask(
                    encoder_attention_mask
                )
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            query_length=query_length,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = (
            self.pooler(sequence_output) if self.pooler is not None else None
        )

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )


class BertLMHeadModel(BertPreTrainedModel):

    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids", r"predictions.decoder.bias"]

    def __init__(self, config):
        super().__init__(config)

        self.bert = BertModel(config, add_pooling_layer=False)
        self.cls = BertOnlyMLMHead(config)

        self.init_weights()

    def get_output_embeddings(self):
        return self.cls.predictions.decoder

    def set_output_embeddings(self, new_embeddings):
        self.cls.predictions.decoder = new_embeddings

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        position_ids=None,
        head_mask=None,
        query_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        labels=None,
        past_key_values=None,
        use_cache=True,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        return_logits=False,
        is_decoder=True,
        reduction="mean",
    ):
        r"""
        encoder_hidden_states  (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length, hidden_size)`, `optional`):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in ``[0, 1]``:
            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the left-to-right language modeling loss (next word prediction). Indices should be in
            ``[-100, 0, ..., config.vocab_size]`` (see ``input_ids`` docstring) Tokens with indices set to ``-100`` are
            ignored (masked), the loss is only computed for the tokens with labels n ``[0, ..., config.vocab_size]``
        past_key_values (:obj:`tuple(tuple(torch.FloatTensor))` of length :obj:`config.n_layers` with each tuple having 4 tensors of shape :obj:`(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.
            If :obj:`past_key_values` are used, the user can optionally input only the last :obj:`decoder_input_ids`
            (those that don't have their past key value states given to this model) of shape :obj:`(batch_size, 1)`
            instead of all :obj:`decoder_input_ids` of shape :obj:`(batch_size, sequence_length)`.
        use_cache (:obj:`bool`, `optional`):
            If set to :obj:`True`, :obj:`past_key_values` key value states are returned and can be used to speed up
            decoding (see :obj:`past_key_values`).
        Returns:
        Example::
            >>> from transformers import BertTokenizer, BertLMHeadModel, BertConfig
            >>> import torch
            >>> tokenizer = BertTokenizer.from_pretrained('bert-base-cased')
            >>> config = BertConfig.from_pretrained("bert-base-cased")
            >>> model = BertLMHeadModel.from_pretrained('bert-base-cased', config=config)
            >>> inputs = tokenizer("Hello, my dog is cute", return_tensors="pt")
            >>> outputs = model(**inputs)
            >>> prediction_logits = outputs.logits
        """
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )
        if labels is not None:
            use_cache = False
        if past_key_values is not None:
            query_embeds = None

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            query_embeds=query_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            is_decoder=is_decoder,
        )

        sequence_output = outputs[0]
        if query_embeds is not None:
            sequence_output = outputs[0][:, query_embeds.shape[1] :, :]

        prediction_scores = self.cls(sequence_output)

        if return_logits:
            return prediction_scores[:, :-1, :].contiguous()

        lm_loss = None
        if labels is not None:
            # we are doing next-token prediction; shift prediction scores and input ids by one
            shifted_prediction_scores = prediction_scores[:, :-1, :].contiguous()
            labels = labels[:, 1:].contiguous()
            loss_fct = CrossEntropyLoss(reduction=reduction, label_smoothing=0.1)
            lm_loss = loss_fct(
                shifted_prediction_scores.view(-1, self.config.vocab_size),
                labels.view(-1),
            )
            if reduction == "none":
                lm_loss = lm_loss.view(prediction_scores.size(0), -1).sum(1)

        if not return_dict:
            output = (prediction_scores,) + outputs[2:]
            return ((lm_loss,) + output) if lm_loss is not None else output

        return CausalLMOutputWithCrossAttentions(
            loss=lm_loss,
            logits=prediction_scores,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
            cross_attentions=outputs.cross_attentions,
        )

    def prepare_inputs_for_generation(
        self, input_ids, query_embeds, past=None, attention_mask=None, **model_kwargs
    ):
        # if model is used as a decoder in encoder-decoder model, the decoder attention mask is created on the fly
        if attention_mask is None:
            attention_mask = input_ids.new_ones(input_ids.shape)
        query_mask = input_ids.new_ones(query_embeds.shape[:-1])
        attention_mask = torch.cat([query_mask, attention_mask], dim=-1)

        # cut decoder_input_ids if past is used
        if past is not None:
            input_ids = input_ids[:, -1:]

        return {
            "input_ids": input_ids,
            "query_embeds": query_embeds,
            "attention_mask": attention_mask,
            "past_key_values": past,
            "encoder_hidden_states": model_kwargs.get("encoder_hidden_states", None),
            "encoder_attention_mask": model_kwargs.get("encoder_attention_mask", None),
            "is_decoder": True,
        }

    def _reorder_cache(self, past, beam_idx):
        reordered_past = ()
        for layer_past in past:
            reordered_past += (
                tuple(
                    past_state.index_select(0, beam_idx) for past_state in layer_past
                ),
            )
        return reordered_past


class BertForMaskedLM(BertPreTrainedModel):

    _keys_to_ignore_on_load_unexpected = [r"pooler"]
    _keys_to_ignore_on_load_missing = [r"position_ids", r"predictions.decoder.bias"]

    def __init__(self, config):
        super().__init__(config)

        self.bert = BertModel(config, add_pooling_layer=False)
        self.cls = BertOnlyMLMHead(config)

        self.init_weights()

    def get_output_embeddings(self):
        return self.cls.predictions.decoder

    def set_output_embeddings(self, new_embeddings):
        self.cls.predictions.decoder = new_embeddings

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        position_ids=None,
        head_mask=None,
        query_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
        return_logits=False,
        is_decoder=False,
    ):
        r"""
        labels (:obj:`torch.LongTensor` of shape :obj:`(batch_size, sequence_length)`, `optional`):
            Labels for computing the masked language modeling loss. Indices should be in ``[-100, 0, ...,
            config.vocab_size]`` (see ``input_ids`` docstring) Tokens with indices set to ``-100`` are ignored
            (masked), the loss is only computed for the tokens with labels in ``[0, ..., config.vocab_size]``
        """

        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            query_embeds=query_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            is_decoder=is_decoder,
        )

        if query_embeds is not None:
            sequence_output = outputs[0][:, query_embeds.shape[1] :, :]
        prediction_scores = self.cls(sequence_output)

        if return_logits:
            return prediction_scores

        masked_lm_loss = None
        if labels is not None:
            loss_fct = CrossEntropyLoss()  # -100 index = padding token
            masked_lm_loss = loss_fct(
                prediction_scores.view(-1, self.config.vocab_size), labels.view(-1)
            )

        if not return_dict:
            output = (prediction_scores,) + outputs[2:]
            return (
                ((masked_lm_loss,) + output) if masked_lm_loss is not None else output
            )

        return MaskedLMOutput(
            loss=masked_lm_loss,
            logits=prediction_scores,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



"""Tests for tonal_dynamic_range module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_tonal_dynamic_range_basics():
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule
    _test_module_basics(TonalDynamicRangeModule, "tonal_dynamic_range")

def test_tonal_dynamic_range_image(image_sample):
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule
    image_sample.quality_metrics = QualityMetrics()
    m = TonalDynamicRangeModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_tonal_dynamic_range_video(video_sample):
    from ayase.modules.tonal_dynamic_range import TonalDynamicRangeModule
    video_sample.quality_metrics = QualityMetrics()
    m = TonalDynamicRangeModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""SQI — Streaming Quality Index (2016). sqi_score — higher = better"""
import logging
from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule
logger = logging.getLogger(__name__)
class SQIModule(PipelineModule):
    name = "sqi"; description = "SQI streaming quality index (2016)"; default_config = {}
    def process(self, sample):
        """Heuristic: base_quality * stalling_factor. Without network data, proxy from metadata."""
        if not sample.is_video: return sample
        try:
            vm = sample.video_metadata
            if vm is None: return sample
            # Base quality from resolution + bitrate
            pixels = vm.width * vm.height
            res_q = min(pixels / (1920*1080), 1.0)
            br_q = min((vm.bitrate or 5_000_000) / 10_000_000, 1.0)
            fps_q = min(vm.fps / 30.0, 1.0) if vm.fps > 0 else 0.5
            base = 0.4*res_q + 0.4*br_q + 0.2*fps_q
            # No stalling data available from file, assume no stalls
            stalling_factor = 1.0
            score = base * stalling_factor
            if sample.quality_metrics is None: sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.sqi_score = float(min(max(score, 0), 1))
        except Exception as e: logger.warning(f"SQI failed: {e}")
        return sample



from typing import List, Union
from PIL import Image

import torch

from transformers.feature_extraction_utils import BatchFeature
from transformers.image_utils import ImageInput, get_image_size, to_numpy_array
from transformers.processing_utils import ProcessingKwargs, ProcessorMixin, Unpack, _validate_images_text_input_order
from transformers.tokenization_utils_base import PreTokenizedInput, TextInput
from transformers.utils import logging
from transformers import Qwen2VLImageProcessor
from transformers.models.qwen2_vl.image_processing_qwen2_vl import smart_resize

logger = logging.get_logger(__name__)


class TarsierProcessorKwargs(ProcessingKwargs, total=False):
    _defaults = {
        "text_kwargs": {},
        "images_kwargs": {},
    }


class TarsierProcessor(ProcessorMixin):

    attributes = ["image_processor", "tokenizer"]
    valid_kwargs = ["chat_template", "image_token", "patch_size", "merge_size", "temporal_patch_size", "max_seq_len"]
    image_processor_class = "AutoImageProcessor"
    tokenizer_class = "AutoTokenizer"

    def __init__(
                self,
                image_processor=None,
                tokenizer=None,
                chat_template=None,
                image_token="<image>",
                patch_size=None,
                merge_size=1,
                temporal_patch_size=1,
                max_seq_len=8192,
                **kwargs,
            ) -> None:
        
        self.image_token = image_token
        self.patch_size = patch_size
        self.merge_size = merge_size
        self.temporal_patch_size = temporal_patch_size
        self.max_seq_len = max_seq_len
        self.max_pixels_per_sample = 128 * 384 * 384

        super().__init__(image_processor, tokenizer, chat_template=chat_template)

    def __call__(
            self, 
            messages,
            image_processing_config=None,
            is_training=True,
        ) -> torch.Tensor:

        output_kwargs = self._merge_kwargs(
            TarsierProcessorKwargs,
            tokenizer_init_kwargs=self.tokenizer.init_kwargs,
        )

        # 【图片处理】
        pixel_values, image_grid_thw = [], []
        num_images = 0
        for msg in messages:
            for content in msg['content']:
                if content['type'] == 'image':
                    num_images += self.temporal_patch_size
                elif content['type'] == 'video':
                    num_images += len(content['video'])
        if num_images > 0 and self.max_pixels_per_sample // num_images < image_processing_config['max_pixels']:
            image_processing_config['max_pixels'] = self.max_pixels_per_sample // num_images
            image_processing_config['min_pixels'] = min(image_processing_config['min_pixels'], image_processing_config['max_pixels'])

        for msg in messages:
            for content in msg['content']:
                if content['type'] == 'image':
                    content['image'] = self.preprocess_image(content['image'], image_processing_config)
                    content['image'] = self.image_processor(images = content['image'], **output_kwargs["images_kwargs"], return_tensors="pt")
                    content['num_vision_tokens'] = self.get_num_vision_tokens(content)
                    pixel_values.append(content['image']['pixel_values'])
                    if 'image_grid_thw' in content['image']:
                        image_grid_thw.extend(content['image']['image_grid_thw'])
                elif content['type'] == 'video':
                    content['video'] = self.preprocess_image(content['video'], image_processing_config)
                    if isinstance(self.image_processor, Qwen2VLImageProcessor):
                        content['video'] = self.image_processor(images = None, videos = content['video'], **output_kwargs["images_kwargs"], return_tensors="pt")
                        pixel_values.append(content['video']['pixel_values_videos'])
                    else:
                        content['video'] = self.image_processor(images = content['video'], **output_kwargs["images_kwargs"], return_tensors="pt")
                        pixel_values.append(content['video']['pixel_values'])

                    if 'video_grid_thw' in content['video']:
                        image_grid_thw.extend(content['video']['video_grid_thw'])
                    content['num_vision_tokens'] = self.get_num_vision_tokens(content)
        
        #【文本处理】
        add_generation_prompt = (not is_training and messages[-1]['role'] != 'assistant')
        strip_final_eos = (not is_training and messages[-1]['role'] == 'assistant')
        text_inputs = self.tokenizer.apply_chat_template(
            messages,
            chat_template = self.chat_template,
            tokenize=True,
            tokenizer_kwargs = output_kwargs["text_kwargs"],
            return_assistant_tokens_mask=True, 
            return_dict=True,
            add_generation_prompt=add_generation_prompt,    
            strip_final_eos=strip_final_eos,    
        )
        labels = [-100 if j == 0 else i for i, j in zip(text_inputs['input_ids'], text_inputs['assistant_masks'])]
        labels = labels[:self.max_seq_len]
        input_ids = text_inputs['input_ids'][:self.max_seq_len]

        image_token_id = self.tokenizer.convert_tokens_to_ids(self.image_token)
        if image_token_id in text_inputs['input_ids'][self.max_seq_len:]:
            raise ValueError(f'Too long sequence! {len(text_inputs["input_ids"])}')
        
        outputs = {
            'input_ids': input_ids,
            'labels': labels,
            'num_images': num_images,
        }
        if len(pixel_values) > 0:
            outputs['pixel_values'] = torch.cat(pixel_values, dim=0)
        if len(image_grid_thw) > 0:
            outputs['image_grid_thw'] = torch.stack(image_grid_thw)
        return outputs
    

    def preprocess_image(self, pil_img: Union[Image.Image, List[Image.Image]], image_processing_config):
        if image_processing_config is None:
            return pil_img
        images = pil_img
        if isinstance(pil_img, Image.Image):
            images = [images]
        if image_processing_config['do_crop']:
            images = [self.centralcrop(img, rate=[4, 3]) for img in images]
        if image_processing_config['do_padding']:
            images = [self.expand2square(
                img,
                # tuple(int(x * 255) for x in self.processor.image_processor.image_mean)
                tuple(int(x * 255) for x in [0, 0, 0])
            ) for img in images]
        if image_processing_config['do_resize']:
            images = [self.resize2square(img) for img in images]
        if image_processing_config.get('max_pixels'):
            images = [self.resize2pixels(
                img, 
                int(image_processing_config['max_pixels']), 
                int(image_processing_config['min_pixels'])
            ) for img in images]
        if isinstance(pil_img, Image.Image):
            images = images[0]
        return images

    def expand2square(self, pil_img, background_color):
        width, height = pil_img.size
        if width == height:
            return pil_img
        elif width > height:
            result = Image.new(pil_img.mode, (width, width), background_color)
            result.paste(pil_img, (0, (width - height) // 2))
            return result
        else:
            result = Image.new(pil_img.mode, (height, height), background_color)
            result.paste(pil_img, ((height - width) // 2, 0))
            return result

    def resize2square(self, pil_img: Image.Image):
        width, height = pil_img.size
        pil_img = pil_img.resize((max(width, height), max(width, height)))
        return pil_img
    
    def centralcrop(self, pil_img: Image.Image, rate=[4, 3]):
        width, height = pil_img.size
        size = (width, height)
        min_len = min(size)
        longer_side = 0 if width >= height else 1
        center = (width/2, height/2)
        box = [0, 0, size[0], size[1]]

        # if longer_side == 0:
        #     box[0] = max(0, center[0] - 1/2*min_len/rate[1]*rate[0])
        #     box[2] = min(width, center[0] + 1/2*min_len/rate[1]*rate[0])
        # else:
        #     box[1] = max(0, center[1] - 1/2*min_len/rate[1]*rate[0])
        #     box[3] = min(height, center[1] + 1/2*min_len/rate[1]*rate[0])
        box[longer_side] = max(0, center[longer_side] - 1/2*min_len/rate[1]*rate[0])
        box[2 + longer_side] = min(size[longer_side], center[longer_side] + 1/2*min_len/rate[1]*rate[0])

        # box = (width/2-min_len/2, height/2-min_len/2, width/2+min_len/2, height/2+min_len/2)
        pil_img = pil_img.crop(box)
        return pil_img
    
    def resize2pixels(self, pil_img: Image.Image, max_pixels=None, min_pixels=None):
        width, height = pil_img.size
        new_height, new_width = smart_resize(height, width, factor=1, max_pixels=max_pixels, min_pixels=min_pixels)
        pil_img = pil_img.resize((new_width, new_height))
        return pil_img

    def get_num_vision_tokens(self, content):
        if isinstance(self.image_processor, Qwen2VLImageProcessor):
            merge_length = self.image_processor.merge_size**2
            if content['type'] == 'image':
                num_image_tokens = content['image']['image_grid_thw'].prod() // merge_length
            else:
                num_image_tokens = content['video']['video_grid_thw'].prod() // merge_length
            return num_image_tokens
        else:
            # 其他模型：image tokens (-> 2x2 compressed) -> add image_newline and image_new
            k = 'image'if content['type'] == 'image' else 'video'
            pixel_values = content[k]['pixel_values'][0]
            n_frames = len(content[k]['pixel_values'])
                
            height, width = get_image_size(to_numpy_array(pixel_values))
            num_image_tokens = (height // (self.patch_size * self.merge_size)) * (width // (self.patch_size * self.merge_size) + 1) + 1
            return num_image_tokens * n_frames
      
    def batch_decode(self, *args, **kwargs):
        """
        This method forwards all its arguments to Qwen2TokenizerFast's [`~PreTrainedTokenizer.batch_decode`]. Please
        refer to the docstring of this method for more information.
        """
        return self.tokenizer.batch_decode(*args, **kwargs)

    def decode(self, *args, **kwargs):
        """
        This method forwards all its arguments to Qwen2TokenizerFast's [`~PreTrainedTokenizer.decode`]. Please refer to
        the docstring of this method for more information.
        """
        return self.tokenizer.decode(*args, **kwargs)

    @property
    def model_input_names(self):
        tokenizer_input_names = self.tokenizer.model_input_names
        image_processor_input_names = self.image_processor.model_input_names
        return list(dict.fromkeys(tokenizer_input_names + image_processor_input_names))



"""TRAJAN — trajectory-based motion consistency assessment.

Tracks feature points across video frames and measures trajectory
smoothness to assess motion quality.

Backend tiers:
  1. **CoTracker** — Facebook's dense point tracking model
     (``pip install cotracker``, ``github.com/facebookresearch/co-tracker``)
  2. **Lucas-Kanade** — OpenCV sparse optical flow tracking (heuristic)
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class TRAJANModule(PipelineModule):
    name = "trajan"
    description = "Motion consistency via point tracking (CoTracker or Lucas-Kanade fallback)"
    default_config = {
        "num_frames": 16,
        "num_points": 256,
    }

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._backend = "lk"
        self._cotracker = None
        self._device = None

    def setup(self) -> None:
        # Tier 1: CoTracker from Facebook Research
        try:
            import torch
            import cotracker

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            # Load CoTracker model
            self._cotracker = torch.hub.load("facebookresearch/co-tracker", "cotracker2").to(device).eval()
            self._device = device
            self._backend = "cotracker"
            self._ml_available = True
            logger.info("TRAJAN loaded CoTracker on %s", device)
            return
        except (ImportError, Exception) as e:
            logger.info("CoTracker unavailable: %s", e)

        # Tier 1b: Try torch.hub directly without cotracker package
        try:
            import torch

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self._cotracker = torch.hub.load("facebookresearch/co-tracker", "cotracker2").to(device).eval()
            self._device = device
            self._backend = "cotracker"
            self._ml_available = True
            logger.info("TRAJAN loaded CoTracker via torch.hub on %s", device)
            return
        except (ImportError, Exception) as e:
            logger.info("CoTracker torch.hub unavailable: %s", e)

        # Tier 2: Lucas-Kanade (OpenCV)
        try:
            import cv2
            self._ml_available = True
            self._backend = "lk"
            logger.info("TRAJAN using Lucas-Kanade optical flow tracking")
        except ImportError:
            logger.warning("TRAJAN unavailable: OpenCV not installed")

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not self._ml_available:
            return sample
        if not sample.is_video:
            return sample

        try:
            score = self._compute_trajectory_consistency(sample)
            if score is not None:
                sample.quality_metrics.trajan_score = score
        except Exception as e:
            logger.warning("TRAJAN processing failed: %s", e)
        return sample

    def _compute_trajectory_consistency(self, sample: Sample) -> Optional[float]:
        """Compute trajectory consistency using the best available backend."""
        if self._backend == "cotracker":
            try:
                return self._compute_cotracker(sample)
            except Exception as e:
                logger.info("CoTracker failed: %s, falling back to LK", e)

        return self._compute_lk(sample)

    def _compute_cotracker(self, sample: Sample) -> Optional[float]:
        """Compute trajectory consistency using CoTracker."""
        import torch
        import cv2

        num_frames = self.config.get("num_frames", 16)
        cap = cv2.VideoCapture(str(sample.path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]

        frames = []
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                frames.append(rgb)
        cap.release()

        if len(frames) < 3:
            return None

        # Prepare video tensor: [1, T, 3, H, W]
        # Resize to manageable size for CoTracker
        h, w = frames[0].shape[:2]
        target_h = min(h, 384)
        target_w = min(w, 512)
        resized = [cv2.resize(f, (target_w, target_h)) for f in frames]

        video_tensor = torch.from_numpy(np.stack(resized)).permute(0, 3, 1, 2).unsqueeze(0).float().to(self._device)

        with torch.no_grad():
            # CoTracker returns (predictions, visibility)
            # predictions shape: [1, T, N, 2] — N tracked points across T frames
            pred_tracks, pred_visibility = self._cotracker(video_tensor)

        tracks = pred_tracks[0].cpu().numpy()  # [T, N, 2]
        visibility = pred_visibility[0].cpu().numpy()  # [T, N]

        # Filter to well-tracked points (visible in most frames)
        min_visible = len(frames) * 0.6
        good_points = visibility.sum(axis=0) >= min_visible
        if good_points.sum() < 5:
            return None

        tracks = tracks[:, good_points, :]  # [T, N_good, 2]

        # Compute trajectory smoothness from CoTracker output
        velocities = np.diff(tracks, axis=0)  # [T-1, N, 2]
        accelerations = np.diff(velocities, axis=0)  # [T-2, N, 2]

        accel_mag = np.sqrt(np.sum(accelerations ** 2, axis=-1))
        vel_mag = np.sqrt(np.sum(velocities[:-1] ** 2, axis=-1))
        vel_mag = np.maximum(vel_mag, 1e-6)

        jerk_ratio = accel_mag / vel_mag
        mean_jerk = float(np.mean(jerk_ratio))
        smoothness = 1.0 / (1.0 + mean_jerk)

        return float(smoothness)

    def _compute_lk(self, sample: Sample) -> Optional[float]:
        """Compute trajectory consistency using Lucas-Kanade optical flow."""
        import cv2

        num_frames = self.config.get("num_frames", 16)
        num_points = self.config.get("num_points", 256)

        cap = cv2.VideoCapture(str(sample.path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]

        frames = []
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
                frames.append(gray)
        cap.release()

        if len(frames) < 3:
            return None

        feature_params = dict(maxCorners=num_points, qualityLevel=0.01, minDistance=10)
        points = cv2.goodFeaturesToTrack(frames[0], mask=None, **feature_params)
        if points is None or len(points) < 10:
            return None

        lk_params = dict(
            winSize=(21, 21),
            maxLevel=3,
            criteria=(cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 30, 0.01),
        )

        trajectories = [points.reshape(-1, 2)]
        current_points = points

        for i in range(1, len(frames)):
            next_points, status, _ = cv2.calcOpticalFlowPyrLK(
                frames[i - 1], frames[i], current_points, None, **lk_params
            )
            if next_points is None:
                break
            good = status.ravel() == 1
            if good.sum() < 5:
                break
            trajectories.append(next_points[good].reshape(-1, 2))
            current_points = next_points[good].reshape(-1, 1, 2)

        if len(trajectories) < 3:
            return None

        min_pts = min(len(t) for t in trajectories)
        if min_pts < 5:
            return None

        trajectories = [t[:min_pts] for t in trajectories]
        traj_array = np.array(trajectories)  # (T, N, 2)

        velocities = np.diff(traj_array, axis=0)
        accelerations = np.diff(velocities, axis=0)

        accel_mag = np.sqrt(np.sum(accelerations ** 2, axis=-1))
        vel_mag = np.sqrt(np.sum(velocities[:-1] ** 2, axis=-1))
        vel_mag = np.maximum(vel_mag, 1e-6)

        jerk_ratio = accel_mag / vel_mag
        mean_jerk = float(np.mean(jerk_ratio))
        smoothness = 1.0 / (1.0 + mean_jerk)

        return float(smoothness)



"""DOVER (Disentangled Objective Video Quality Evaluator) module.

ICCV 2023 — state-of-the-art no-reference video quality assessment
that disentangles *technical* quality (noise, blur, compression) from
*aesthetic* quality (composition, colour harmony, content).

Supports two backends:
1. Native DOVER — loads the original model directly (``pip install dover``
   or ``pip install git+https://github.com/VQAssessment/DOVER.git``).
2. pyiqa wrapper — ``pip install pyiqa`` (if pyiqa ships DOVER).

dover_score    — fused overall quality (higher = better, 0-1 sigmoid)
dover_technical — technical sub-score (0-1 sigmoid)
dover_aesthetic — aesthetic sub-score (0-1 sigmoid)
"""

import logging
import os
from typing import Optional

import numpy as np

from ayase.models import Sample, QualityMetrics, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _fuse_results(results):
    """Sigmoid rescaling from raw DOVER outputs (matches the original repo)."""
    t = (results[1] + 0.0758) / 0.0129
    a = (results[0] - 0.1253) / 0.0318
    x = t * 0.6104 + a * 0.3896
    return {
        "aesthetic": 1.0 / (1.0 + np.exp(-a)),
        "technical": 1.0 / (1.0 + np.exp(-t)),
        "overall": 1.0 / (1.0 + np.exp(-x)),
    }


class DOVERModule(PipelineModule):
    name = "dover"
    description = "DOVER disentangled technical + aesthetic VQA (ICCV 2023)"
    default_config = {
        "warning_threshold": 0.4,
        "weights_path": None,  # Explicit path to DOVER.pth (optional)
        "preferred_backend": None,  # "native", "onnx", or "pyiqa" — None = auto (native first)
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.warning_threshold = self.config.get("warning_threshold", 0.4)
        self.weights_path = self.config.get("weights_path", None)
        self._device = "cpu"
        self._ml_available = False
        self._backend = None  # "native" or "pyiqa"

        # Native backend state
        self._model = None
        self._sample_types = None
        self._mean = None
        self._std = None

    def setup(self) -> None:
        # In test mode, skip all heavy model loading
        if self.test_mode:
            self._backend = "heuristic"
            logger.debug("DOVER: test mode, skipping ML backends")
            return

        preferred = self.config.get("preferred_backend")

        # If explicit backend requested, try it first
        if preferred == "onnx":
            if self._try_onnx_setup():
                return
        elif preferred == "pyiqa":
            if self._try_pyiqa_setup():
                return
        elif preferred == "native":
            if self._try_native_setup():
                return

        # Auto fallback: native → ONNX → pyiqa
        if not self._ml_available and self._try_native_setup():
            return
        if not self._ml_available and self._try_onnx_setup():
            return
        if not self._ml_available and self._try_pyiqa_setup():
            return

        logger.warning(
            "DOVER unavailable. Install with: "
            "pip install git+https://github.com/VQAssessment/DOVER.git  "
            "OR  pip install pyiqa"
        )

    # ------------------------------------------------------------------ #
    #  Backend 1: Native DOVER                                            #
    # ------------------------------------------------------------------ #

    def _try_native_setup(self) -> bool:
        try:
            import torch
            from ayase.third_party.dover.models import DOVER as DOVERModel
            from ayase.third_party.dover.datasets import UnifiedFrameSampler, spatial_temporal_view_decomposition  # noqa: F401

            self._device = "cuda" if torch.cuda.is_available() else "cpu"

            # Resolve weights path
            weights = self._resolve_weights()
            if weights is None:
                logger.debug("DOVER weights not found, skipping native backend.")
                return False

            # Hardcoded config matching the original val-l1080p profile
            self._sample_types = {
                "technical": {
                    "fragments_h": 7, "fragments_w": 7,
                    "fsize_h": 32, "fsize_w": 32,
                    "aligned": 32,
                    "clip_len": 32, "frame_interval": 2, "num_clips": 3,
                },
                "aesthetic": {
                    "size_h": 224, "size_w": 224,
                    "clip_len": 32, "frame_interval": 2,
                    "t_frag": 32, "num_clips": 1,
                },
            }

            model_args = {
                "backbone": {
                    "technical": {"type": "swin_tiny_grpb", "checkpoint": True},
                    "aesthetic": {"type": "conv_tiny"},
                },
                "backbone_preserve_keys": "technical,aesthetic",
                "divide_head": True,
                "vqa_head": {"in_channels": 768, "hidden_channels": 64},
            }

            # Pre-cache ConvNeXt backbone weights to avoid torch.hub download from fbaipublicfiles
            self._ensure_convnext_cached()

            self._model = DOVERModel(**model_args).to(self._device)
            self._model.load_state_dict(
                torch.load(weights, map_location=self._device, weights_only=True)
            )
            self._model.eval()

            self._mean = torch.FloatTensor([123.675, 116.28, 103.53])
            self._std = torch.FloatTensor([58.395, 57.12, 57.375])

            self._ml_available = True
            self._backend = "native"
            logger.info(f"DOVER (native) initialised on {self._device}")
            return True

        except ImportError:
            logger.debug("dover package not installed, trying pyiqa fallback.")
            return False
        except Exception as e:
            logger.debug(f"Native DOVER setup failed: {e}")
            return False

    # Original: https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth
    _CONVNEXT_URL = "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/dover/convnext_tiny_1k_224_ema.pth"
    _CONVNEXT_FILENAME = "convnext_tiny_1k_224_ema.pth"

    def _ensure_convnext_cached(self) -> None:
        """Download ConvNeXt weights to torch hub cache if not present."""
        import torch
        hub_dir = torch.hub.get_dir()
        cache_dir = os.path.join(hub_dir, "checkpoints")
        cached = os.path.join(cache_dir, self._CONVNEXT_FILENAME)
        if os.path.exists(cached):
            return
        os.makedirs(cache_dir, exist_ok=True)
        logger.info("Downloading ConvNeXt backbone for DOVER...")
        from ayase.config import download_model_file
        tmp = download_model_file(
            os.path.join("hub", "checkpoints", self._CONVNEXT_FILENAME),
            self._CONVNEXT_URL,
            os.path.dirname(hub_dir),  # parent of hub dir
        )
        # Move to torch cache if downloaded elsewhere
        if str(tmp) != cached and os.path.exists(str(tmp)):
            import shutil
            shutil.copy2(str(tmp), cached)

    # Original: https://github.com/VQAssessment/DOVER/releases/download/v0.1.0/DOVER.pth
    _DOVER_WEIGHTS_URL = "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/dover/DOVER.pth"

    def _resolve_weights(self) -> Optional[str]:
        """Find DOVER.pth weights file, auto-downloading if needed."""
        if self.weights_path and os.path.exists(self.weights_path):
            return self.weights_path

        models_dir = self.config.get("models_dir", "models")

        candidates = [
            os.path.join(models_dir, "dover", "DOVER.pth"),
            os.path.join(models_dir, "DOVER", "pretrained_weights", "DOVER.pth"),
            os.path.join(models_dir, "DOVER.pth"),
        ]

        # Also check TORCH_HOME
        torch_home = os.environ.get("TORCH_HOME", "")
        if torch_home:
            candidates.append(
                os.path.join(torch_home, "DOVER", "pretrained_weights", "DOVER.pth")
            )

        for path in candidates:
            if os.path.exists(path):
                return path

        # Auto-download
        from ayase.config import download_model_file

        return download_model_file(
            "dover/DOVER.pth", self._DOVER_WEIGHTS_URL, models_dir
        )

    # ------------------------------------------------------------------ #
    #  Backend 2: ONNX                                                    #
    # ------------------------------------------------------------------ #

    # Original: https://github.com/VQAssessment/DOVER (convert_to_onnx.py)
    _ONNX_WEIGHTS_URL = "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/dover/onnx_dover.onnx"

    def _try_onnx_setup(self) -> bool:
        try:
            import onnxruntime as ort

            models_dir = self.config.get("models_dir", "models")
            onnx_path = os.path.join(models_dir, "dover", "onnx_dover.onnx")

            if not os.path.exists(onnx_path):
                from ayase.config import download_model_file
                try:
                    download_model_file("dover/onnx_dover.onnx", self._ONNX_WEIGHTS_URL, models_dir)
                except Exception:
                    return False

            if not os.path.exists(onnx_path):
                return False

            import torch
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"] if torch.cuda.is_available() else ["CPUExecutionProvider"]
            self._onnx_session = ort.InferenceSession(onnx_path, providers=providers)
            self._device = "cuda" if torch.cuda.is_available() else "cpu"

            self._sample_types = {
                "technical": {
                    "fragments_h": 7, "fragments_w": 7,
                    "fsize_h": 32, "fsize_w": 32,
                    "aligned": 32,
                    "clip_len": 32, "frame_interval": 2, "num_clips": 3,
                },
                "aesthetic": {
                    "size_h": 224, "size_w": 224,
                    "clip_len": 32, "frame_interval": 2,
                    "t_frag": 32, "num_clips": 1,
                },
            }

            import torch
            self._mean = torch.FloatTensor([123.675, 116.28, 103.53])
            self._std = torch.FloatTensor([58.395, 57.12, 57.375])

            self._ml_available = True
            self._backend = "onnx"
            logger.info("DOVER (ONNX) initialised")
            return True

        except ImportError:
            logger.debug("onnxruntime not installed, skipping ONNX backend.")
            return False
        except Exception as e:
            logger.debug(f"ONNX DOVER setup failed: {e}")
            return False

    # ------------------------------------------------------------------ #
    #  Backend 3: pyiqa                                                   #
    # ------------------------------------------------------------------ #

    def _try_pyiqa_setup(self) -> bool:
        try:
            import pyiqa
            import torch

            self._device = "cuda" if torch.cuda.is_available() else "cpu"

            models_dir = self.config.get("models_dir", None)
            old_torch_home = os.environ.get("TORCH_HOME")
            if models_dir:
                os.environ["TORCH_HOME"] = str(models_dir)

            try:
                self._model = pyiqa.create_metric("dover", device=self._device, as_loss=False)
                self._ml_available = True
                self._backend = "pyiqa"
                logger.info(f"DOVER (pyiqa) initialised on {self._device}")
                return True
            finally:
                if old_torch_home is not None:
                    os.environ["TORCH_HOME"] = old_torch_home
                else:
                    os.environ.pop("TORCH_HOME", None)

        except ImportError:
            return False
        except Exception as e:
            logger.debug(f"pyiqa DOVER setup failed: {e}")
            return False

    # ------------------------------------------------------------------ #
    #  Processing                                                         #
    # ------------------------------------------------------------------ #

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available or not sample.is_video:
            return sample

        try:
            if self._backend == "native":
                aesthetic, technical, overall = self._process_native(sample)
            elif self._backend == "onnx":
                aesthetic, technical, overall = self._process_onnx(sample)
            else:
                aesthetic, technical, overall = self._process_pyiqa(sample)

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.dover_score = overall
            sample.quality_metrics.dover_aesthetic = aesthetic
            sample.quality_metrics.dover_technical = technical

            if overall < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low DOVER score: {overall:.3f}",
                        details={
                            "dover_score": overall,
                            "dover_technical": technical,
                            "dover_aesthetic": aesthetic,
                        },
                        recommendation=(
                            "Video quality is below threshold. "
                            "Technical issues: noise/blur/compression. "
                            "Aesthetic issues: composition/colour."
                        ),
                    )
                )

            logger.debug(
                f"DOVER for {sample.path.name}: "
                f"overall={overall:.3f} tech={technical:.3f} aes={aesthetic:.3f}"
            )

        except Exception as e:
            logger.warning(f"DOVER failed for {sample.path}: {e}")

        return sample

    def _process_native(self, sample: Sample):
        """Run inference using the native DOVER model."""
        import torch
        from ayase.third_party.dover.datasets import UnifiedFrameSampler, spatial_temporal_view_decomposition

        # Build temporal samplers
        temporal_samplers = {}
        for stype, sopt in self._sample_types.items():
            if "t_frag" not in sopt:
                temporal_samplers[stype] = UnifiedFrameSampler(
                    sopt["clip_len"], sopt["num_clips"], sopt["frame_interval"]
                )
            else:
                temporal_samplers[stype] = UnifiedFrameSampler(
                    sopt["clip_len"] // sopt["t_frag"],
                    sopt["t_frag"],
                    sopt["frame_interval"],
                    sopt["num_clips"],
                )

        # Decompose video into views
        views, _ = spatial_temporal_view_decomposition(
            str(sample.path), self._sample_types, temporal_samplers
        )

        # Normalize and reshape
        processed = {}
        for k, v in views.items():
            num_clips = self._sample_types[k].get("num_clips", 1)
            processed[k] = (
                ((v.permute(1, 2, 3, 0) - self._mean) / self._std)
                .permute(3, 0, 1, 2)
                .reshape(v.shape[0], num_clips, -1, *v.shape[2:])
                .transpose(0, 1)
                .to(self._device)
            )

        # Ensure aesthetic-first order for _fuse_results:
        # DOVERModel iterates dict keys, so order depends on self._sample_types.
        # _fuse_results expects [aesthetic, technical].
        ordered = {"aesthetic": processed["aesthetic"], "technical": processed["technical"]}
        with torch.no_grad():
            outputs = self._model(ordered, reduce_scores=False)
            results = [float(np.mean(out.detach().cpu().numpy())) for out in outputs]

        rescaled = _fuse_results(results)
        return rescaled["aesthetic"], rescaled["technical"], rescaled["overall"]

    def _process_onnx(self, sample: Sample):
        """Run inference using ONNX Runtime."""
        from ayase.third_party.dover.datasets import UnifiedFrameSampler, spatial_temporal_view_decomposition

        temporal_samplers = {}
        for stype, sopt in self._sample_types.items():
            if "t_frag" not in sopt:
                temporal_samplers[stype] = UnifiedFrameSampler(
                    sopt["clip_len"], sopt["num_clips"], sopt["frame_interval"]
                )
            else:
                temporal_samplers[stype] = UnifiedFrameSampler(
                    sopt["clip_len"] // sopt["t_frag"],
                    sopt["t_frag"],
                    sopt["frame_interval"],
                    sopt["num_clips"],
                )

        views, _ = spatial_temporal_view_decomposition(
            str(sample.path), self._sample_types, temporal_samplers
        )

        processed = {}
        for k, v in views.items():
            num_clips = self._sample_types[k].get("num_clips", 1)
            processed[k] = (
                ((v.permute(1, 2, 3, 0) - self._mean) / self._std)
                .permute(3, 0, 1, 2)
                .reshape(v.shape[0], num_clips, -1, *v.shape[2:])
                .transpose(0, 1)
            )

        preds = self._onnx_session.run(
            None,
            {
                "aes_view": processed["aesthetic"].numpy(),
                "tech_view": processed["technical"].numpy(),
            },
        )
        results = [float(np.mean(pred)) for pred in preds]

        rescaled = _fuse_results(results)
        return rescaled["aesthetic"], rescaled["technical"], rescaled["overall"]

    def _process_pyiqa(self, sample: Sample):
        """Run inference using pyiqa's DOVER wrapper."""
        import torch

        with torch.no_grad():
            result = self._model(str(sample.path))

        if isinstance(result, torch.Tensor):
            result = result.squeeze().cpu().tolist()

        if isinstance(result, (list, tuple)):
            if len(result) >= 2:
                aesthetic = float(result[0])
                technical = float(result[1])
                overall = (aesthetic + technical) / 2.0
            else:
                aesthetic = technical = overall = float(result[0])
        else:
            aesthetic = technical = overall = float(result)

        return aesthetic, technical, overall



"""
 Copyright (c) 2023, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""
import logging

import torch
import torch.distributed as dist
import torch.nn as nn
from torch.cuda.amp import autocast as autocast
from torch.nn import functional as F

from ...common.registry import registry
from ..base_model import all_gather_with_grad, concat_all_gather
from .blip2 import (
    Blip2Base,
    compute_sim_matrix,
    disabled_train,
)
from ..blip_models.blip_outputs import BlipOutput, BlipOutputFeatures


@registry.register_model("blip2")
@registry.register_model("blip2_feature_extractor")
class Blip2Qformer(Blip2Base):
    """
    BLIP2 stage-1 model with Q-former and ViT.
    Supported model types:
        - pretrained: pretrained model with vit-g
        - pretrain_vitL: pretrained model with vit-large
        - coco: fintuned model on coco
    Usage:
        >>> from lavis.models import load_model
        >>> model = load_model("blip2", "pretrain")
    """

    PRETRAINED_MODEL_CONFIG_DICT = {
        "pretrain": "configs/models/blip2/blip2_pretrain.yaml",
        "pretrain_vitL": "configs/models/blip2/blip2_pretrain_vitL.yaml",
        "coco": "configs/models/blip2/blip2_coco.yaml",
    }

    def __init__(
        self,
        vit_model="eva_clip_g",
        img_size=224,
        drop_path_rate=0,
        use_grad_checkpoint=False,
        vit_precision="fp16",
        freeze_vit=True,
        num_query_token=32,
        cross_attention_freq=2,
        embed_dim=256,
        max_txt_len=32,
    ):
        super().__init__()

        self.tokenizer = self.init_tokenizer()

        self.visual_encoder, self.ln_vision = self.init_vision_encoder(
            vit_model, img_size, drop_path_rate, use_grad_checkpoint, vit_precision
        )
        if freeze_vit:
            for name, param in self.visual_encoder.named_parameters():
                param.requires_grad = False
            self.visual_encoder = self.visual_encoder.eval()
            self.visual_encoder.train = disabled_train
            logging.info("freeze vision encoder")
        self.Qformer, self.query_tokens = self.init_Qformer(
            num_query_token, self.visual_encoder.num_features, cross_attention_freq
        )
        self.Qformer.resize_token_embeddings(len(self.tokenizer))
        state_dict = self.Qformer.state_dict()
        for name, param in self.Qformer.named_parameters():
            if "_query" in name:
                key_orig = name.replace("_query", "")
                param.data.copy_(state_dict[key_orig])

        self.vision_proj = nn.Linear(self.Qformer.config.hidden_size, embed_dim)
        self.text_proj = nn.Linear(self.Qformer.config.hidden_size, embed_dim)

        self.itm_head = nn.Linear(self.Qformer.config.hidden_size, 2)

        self.temp = nn.Parameter(0.07 * torch.ones([]))

        self.max_txt_len = max_txt_len

    def forward(self, samples):
        image = samples["image"]
        text = samples["text_input"]

        image_embeds = self.ln_vision(self.visual_encoder(image))
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            image.device
        )

        query_tokens = self.query_tokens.expand(image_embeds.shape[0], -1, -1)

        query_output = self.Qformer.bert(
            query_embeds=query_tokens,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_atts,
            use_cache=True,
            return_dict=True,
        )

        image_feats = F.normalize(
            self.vision_proj(query_output.last_hidden_state), dim=-1
        )

        text_tokens = self.tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=self.max_txt_len,
            return_tensors="pt",
        ).to(image.device)
        text_output = self.Qformer.bert(
            text_tokens.input_ids,
            attention_mask=text_tokens.attention_mask,
            return_dict=True,
        )
        text_feat = F.normalize(
            self.text_proj(text_output.last_hidden_state[:, 0, :]), dim=-1
        )

        ###============== Image-text Contrastive ===================###
        image_feats_all = concat_all_gather(
            image_feats
        )  # [batch_size*num_gpu, num_query_tokens, embed_dim]
        text_feat_all = concat_all_gather(text_feat)  # [batch_size*num_gpu, embed_dim]

        sim_q2t = torch.matmul(
            image_feats.unsqueeze(1), text_feat_all.unsqueeze(-1)
        ).squeeze()
        # [batch_size, batch_size*num_gpu, num_query_tokens]

        # image-text similarity: aggregate across all query tokens
        sim_i2t, _ = sim_q2t.max(-1)
        sim_i2t = sim_i2t / self.temp

        # text-query similarity: [batch_size, batch_size*num_gpu, num_query_tokens]
        sim_t2q = torch.matmul(
            text_feat.unsqueeze(1).unsqueeze(1), image_feats_all.permute(0, 2, 1)
        ).squeeze()

        # text-image similarity: aggregate across all query tokens
        sim_t2i, _ = sim_t2q.max(-1)
        sim_t2i = sim_t2i / self.temp  # [batch_size, batch_size*num_gpu]

        rank = dist.get_rank()
        bs = image.size(0)
        targets = torch.linspace(rank * bs, rank * bs + bs - 1, bs, dtype=int).to(
            image.device
        )

        loss_itc = (
            F.cross_entropy(sim_i2t, targets, label_smoothing=0.1)
            + F.cross_entropy(sim_t2i, targets, label_smoothing=0.1)
        ) / 2

        ###============== Image-text Matching ===================###
        text_input_ids_world = concat_all_gather(text_tokens.input_ids)
        text_attention_mask_world = concat_all_gather(text_tokens.attention_mask)
        image_embeds_world = all_gather_with_grad(image_embeds)
        with torch.no_grad():
            weights_t2i = F.softmax(sim_t2i, dim=1) + 1e-4
            weights_t2i[:, rank * bs : rank * bs + bs].fill_diagonal_(0)
            weights_i2t = F.softmax(sim_i2t, dim=1) + 1e-4
            weights_i2t[:, rank * bs : rank * bs + bs].fill_diagonal_(0)

        # select a negative image for each text
        image_embeds_neg = []
        for b in range(bs):
            neg_idx = torch.multinomial(weights_t2i[b], 1).item()
            image_embeds_neg.append(image_embeds_world[neg_idx])
        image_embeds_neg = torch.stack(image_embeds_neg, dim=0)

        # select a negative text for each image
        text_ids_neg = []
        text_atts_neg = []
        for b in range(bs):
            neg_idx = torch.multinomial(weights_i2t[b], 1).item()
            text_ids_neg.append(text_input_ids_world[neg_idx])
            text_atts_neg.append(text_attention_mask_world[neg_idx])

        text_ids_neg = torch.stack(text_ids_neg, dim=0)
        text_atts_neg = torch.stack(text_atts_neg, dim=0)

        text_ids_all = torch.cat(
            [text_tokens.input_ids, text_tokens.input_ids, text_ids_neg], dim=0
        )  # pos, pos, neg
        text_atts_all = torch.cat(
            [text_tokens.attention_mask, text_tokens.attention_mask, text_atts_neg],
            dim=0,
        )

        query_tokens_itm = self.query_tokens.expand(text_ids_all.shape[0], -1, -1)
        query_atts_itm = torch.ones(query_tokens_itm.size()[:-1], dtype=torch.long).to(
            image.device
        )
        attention_mask_all = torch.cat([query_atts_itm, text_atts_all], dim=1)

        image_embeds_all = torch.cat(
            [image_embeds, image_embeds_neg, image_embeds], dim=0
        )  # pos, neg, pos
        image_atts_all = torch.ones(image_embeds_all.size()[:-1], dtype=torch.long).to(
            image.device
        )

        output_itm = self.Qformer.bert(
            text_ids_all,
            query_embeds=query_tokens_itm,
            attention_mask=attention_mask_all,
            encoder_hidden_states=image_embeds_all,
            encoder_attention_mask=image_atts_all,
            return_dict=True,
        )

        vl_embeddings = output_itm.last_hidden_state[:, : query_tokens_itm.size(1), :]
        vl_output = self.itm_head(vl_embeddings)
        logits = vl_output.mean(dim=1)

        itm_labels = torch.cat(
            [torch.ones(bs, dtype=torch.long), torch.zeros(2 * bs, dtype=torch.long)],
            dim=0,
        ).to(image.device)
        loss_itm = F.cross_entropy(logits, itm_labels)

        ##================= Image Captioning ========================##
        decoder_input_ids = text_tokens.input_ids.clone()
        decoder_input_ids[:, 0] = self.tokenizer.bos_token_id
        labels = decoder_input_ids.masked_fill(
            decoder_input_ids == self.tokenizer.pad_token_id, -100
        )

        query_atts = torch.ones(query_tokens.size()[:-1], dtype=torch.long).to(
            image.device
        )
        attention_mask = torch.cat([query_atts, text_tokens.attention_mask], dim=1)
        lm_output = self.Qformer(
            decoder_input_ids,
            attention_mask=attention_mask,
            past_key_values=query_output.past_key_values,
            return_dict=True,
            labels=labels,
        )

        loss_lm = lm_output.loss

        return BlipOutput(
            loss=loss_itc + loss_itm + loss_lm,
            loss_itc=loss_itc,
            loss_itm=loss_itm,
            loss_lm=loss_lm,
        )

    @torch.no_grad()
    def generate(
        self,
        samples,
        use_nucleus_sampling=False,
        num_beams=3,
        max_length=30,
        min_length=10,
        top_p=0.9,
        repetition_penalty=1.0,
    ):
        """
        Args:
            samples (dict): A dictionary containing the following keys:
                - image (torch.Tensor): A tensor of shape (batch_size, 3, H, W)
            use_nucleus_sampling (bool): Whether to use nucleus sampling. If False, use top-k sampling.
            num_beams (int): Number of beams for beam search. 1 means no beam search.
            max_length (int): The maximum length of the sequence to be generated.
            min_length (int): The minimum length of the sequence to be generated.
            top_p (float): The cumulative probability for nucleus sampling.
            repetition_penalty (float): The parameter for repetition penalty. 1.0 means no penalty.
            num_captions (int): Number of captions to be generated for each image.
        Returns:
            captions (list): A list of strings of length batch_size * num_captions.
        """
        image = samples["image"]
        image_embeds = self.ln_vision(self.visual_encoder(image))

        if not use_nucleus_sampling:
            image_embeds = image_embeds.repeat_interleave(num_beams, dim=0)
        else:
            num_beams = 1
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            image.device
        )

        model_kwargs = {
            "encoder_hidden_states": image_embeds,
            "encoder_attention_mask": image_atts,
        }

        input_ids = (
            torch.LongTensor(image.size(0), 1)
            .fill_(self.tokenizer.bos_token_id)
            .to(image.device)
        )
        query_tokens = self.query_tokens.expand(image_embeds.shape[0], -1, -1)

        outputs = self.Qformer.generate(
            input_ids=input_ids,
            query_embeds=query_tokens,
            max_length=max_length,
            min_length=min_length,
            num_beams=num_beams,
            do_sample=use_nucleus_sampling,
            top_p=top_p,
            eos_token_id=self.tokenizer.sep_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            **model_kwargs
        )
        captions = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)
        return captions

    def forward_image(self, image):
        image_embeds = self.ln_vision(self.visual_encoder(image))
        image_atts = torch.ones(image_embeds.size()[:-1], dtype=torch.long).to(
            image.device
        )

        query_tokens = self.query_tokens.expand(image_embeds.shape[0], -1, -1)

        query_output = self.Qformer.bert(
            query_embeds=query_tokens,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_atts,
            return_dict=True,
        )
        return query_output.last_hidden_state, image_embeds

    def forward_text(self, text_tokens):
        text_output = self.Qformer.bert(
            text_tokens.input_ids,
            attention_mask=text_tokens.attention_mask,
            return_dict=True,
        )
        return text_output.last_hidden_state[:, 0, :]

    def compute_itm(self, image_inputs, text_ids, text_atts):
        image_atts = torch.ones(image_inputs.size()[:-1], dtype=torch.long).to(
            image_inputs.device
        )
        query_tokens = self.query_tokens.expand(image_inputs.shape[0], -1, -1)
        query_atts = torch.ones(query_tokens.size()[:-1], dtype=torch.long).to(
            image_inputs.device
        )
        attention_mask = torch.cat([query_atts, text_atts], dim=1)
        output_itm = self.Qformer.bert(
            text_ids,
            query_embeds=query_tokens,
            attention_mask=attention_mask,
            encoder_hidden_states=image_inputs,
            encoder_attention_mask=image_atts,
            return_dict=True,
        )
        vl_embeddings = output_itm.last_hidden_state[:, : query_tokens.size(1), :]
        itm_logit = self.itm_head(vl_embeddings)
        itm_logit = itm_logit[:, :, 1].mean(dim=1)
        return itm_logit

    @torch.no_grad()
    def extract_features(self, samples, mode="multimodal"):
        """
        Extract features for multimodal or unimodal samples.
        Args:
            samples (dict): A dictionary of samples, containing the following keys:
                - image (torch.Tensor): A tensor of shape (B, C, H, W) containing the image.
                    Raw images should be preprocessed before being passed to feature extractor.
                - text_input (list): A list of strings containing the text, length B.
            mode (str): The mode of feature extraction. Can be either "multimodal", "text" or "image".
                If "multimodal", return image features and multimodal features;
                if "text", return text features;
                if "image", return image features.
                Default: "multimodal".
        Returns:
            BlipOutputFeatures: A BlipOutputFeatures object containing the features.
                See lavis/models/blip_models/blip_outputs.py for more details.
        """
        image = samples.get("image")
        caption = samples.get("text_input")

        # assert mode is one of "image", "text", "multimodal"
        assert mode in [
            "image",
            "text",
            "multimodal",
        ], "mode must be one of 'image', 'text', 'multimodal'"

        # initalize output
        image_embeds, text_embeds, multimodal_embeds = None, None, None
        image_features, text_features = None, None

        if mode == "image":
            assert (
                image is not None
            ), "Image is not provided for mode 'image' or 'multimodal'"
            # return query features
            with self.maybe_autocast():
                image_embeds_frozen = self.ln_vision(self.visual_encoder(image))
            image_embeds_frozen = image_embeds_frozen.float()
            image_atts = torch.ones(
                image_embeds_frozen.size()[:-1], dtype=torch.long
            ).to(self.device)
            query_tokens = self.query_tokens.expand(
                image_embeds_frozen.shape[0], -1, -1
            )

            query_output = self.Qformer.bert(
                query_embeds=query_tokens,
                encoder_hidden_states=image_embeds_frozen,
                encoder_attention_mask=image_atts,
                return_dict=True,
            )
            image_embeds = query_output.last_hidden_state
            image_features = F.normalize(self.vision_proj(image_embeds), dim=-1)

        elif mode == "text":
            assert (
                caption is not None
            ), "text input is None for mode 'text' or 'multimodal'"

            # return text features
            text = self.tokenizer(caption, return_tensors="pt", padding=True).to(
                self.device
            )

            text_output = self.Qformer.bert(
                text.input_ids,
                attention_mask=text.attention_mask,
                return_dict=True,
            )
            text_embeds = text_output.last_hidden_state
            text_features = self.text_proj(text_embeds)
            text_features = F.normalize(text_features, dim=-1)

        elif mode == "multimodal":
            # return multimodel query features
            with self.maybe_autocast():
                image_embeds_frozen = self.ln_vision(self.visual_encoder(image))
            image_embeds_frozen = image_embeds_frozen.float()
            image_atts = torch.ones(
                image_embeds_frozen.size()[:-1], dtype=torch.long
            ).to(self.device)
            query_tokens = self.query_tokens.expand(
                image_embeds_frozen.shape[0], -1, -1
            )
            query_atts = torch.ones(query_tokens.size()[:-1], dtype=torch.long).to(
                self.device
            )

            text = self.tokenizer(caption, return_tensors="pt", padding=True).to(
                self.device
            )
            attention_mask = torch.cat([query_atts, text.attention_mask], dim=1)

            output = self.Qformer.bert(
                text.input_ids,
                query_embeds=query_tokens,
                attention_mask=attention_mask,
                encoder_hidden_states=image_embeds_frozen,
                encoder_attention_mask=image_atts,
                return_dict=True,
            )

            multimodal_embeds = output.last_hidden_state[:, : query_tokens.size(1), :]

        return BlipOutputFeatures(
            image_embeds=image_embeds,
            image_embeds_proj=image_features,
            text_embeds=text_embeds,
            text_embeds_proj=text_features,
            multimodal_embeds=multimodal_embeds,
        )

    @classmethod
    def from_config(cls, cfg):
        vit_model = cfg.get("vit_model", "eva_clip_g")
        img_size = cfg.get("image_size")
        num_query_token = cfg.get("num_query_token")
        cross_attention_freq = cfg.get("cross_attention_freq", 2)

        drop_path_rate = cfg.get("drop_path_rate", 0)
        use_grad_checkpoint = cfg.get("use_grad_checkpoint", False)
        vit_precision = cfg.get("vit_precision", "fp16")
        freeze_vit = cfg.get("freeze_vit", True)

        max_txt_len = cfg.get("max_txt_len", 32)

        model = cls(
            vit_model=vit_model,
            img_size=img_size,
            drop_path_rate=drop_path_rate,
            use_grad_checkpoint=use_grad_checkpoint,
            vit_precision=vit_precision,
            freeze_vit=freeze_vit,
            num_query_token=num_query_token,
            cross_attention_freq=cross_attention_freq,
            max_txt_len=max_txt_len,
        )
        model.load_checkpoint_from_config(cfg)

        return model

    def compute_sim_matrix(self, data_loader, task_cfg):
        """
        Compute similarity i2t, t2i matrix for the given data loader.
        """
        k_test = task_cfg.k_test

        return compute_sim_matrix(model=self, data_loader=data_loader, k_test=k_test)



"""Smoke tests for audio metric modules."""

import numpy as np
import pytest
import soundfile as sf


@pytest.fixture
def synthetic_wav(tmp_path):
    """440 Hz sine wave, 1 second, 16 kHz."""
    sr = 16000
    t = np.linspace(0, 1.0, sr, dtype=np.float32)
    audio = 0.5 * np.sin(2 * np.pi * 440 * t)
    path = tmp_path / "ref.wav"
    sf.write(str(path), audio, sr)
    return path


@pytest.fixture
def degraded_wav(tmp_path):
    """440 Hz sine + noise."""
    sr = 16000
    t = np.linspace(0, 1.0, sr, dtype=np.float32)
    audio = 0.5 * np.sin(2 * np.pi * 440 * t) + 0.05 * np.random.randn(sr).astype(np.float32)
    path = tmp_path / "deg.wav"
    sf.write(str(path), audio, sr)
    return path


class TestAudioSISDR:
    def test_identical_signals(self, synthetic_wav):
        from ayase.modules.audio_si_sdr import AudioSISDRModule

        mod = AudioSISDRModule()
        ref = mod._load_audio(synthetic_wav)
        si_sdr = mod._compute_si_sdr(ref, ref)
        assert si_sdr > 50  # identical → very high

    def test_noisy_signal_positive(self, synthetic_wav, degraded_wav):
        from ayase.modules.audio_si_sdr import AudioSISDRModule

        mod = AudioSISDRModule()
        ref = mod._load_audio(synthetic_wav)
        deg = mod._load_audio(degraded_wav)
        si_sdr = mod._compute_si_sdr(ref, deg)
        assert si_sdr > 0  # signal still dominates noise


class TestAudioMCD:
    def test_identical_zero(self, synthetic_wav):
        from ayase.modules.audio_mcd import AudioMCDModule

        mod = AudioMCDModule()
        mod._ml_available = True
        mfcc = mod._extract_mfcc(synthetic_wav)
        assert mfcc is not None
        assert mfcc.shape[0] == 14  # n_mfcc + 1

    def test_self_mcd_near_zero(self, synthetic_wav):
        from ayase.modules.audio_mcd import AudioMCDModule

        mod = AudioMCDModule()
        mod._ml_available = True
        mfcc = mod._extract_mfcc(synthetic_wav)
        diff = mfcc[1:, :] - mfcc[1:, :]
        frame_dist = np.sqrt(np.sum(diff ** 2, axis=0))
        mcd = (10.0 * np.sqrt(2.0) / np.log(10.0)) * np.mean(frame_dist)
        assert mcd == 0.0


class TestAudioLPDist:
    def test_self_distance_zero(self, synthetic_wav):
        from ayase.modules.audio_lpdist import AudioLPDistModule

        mod = AudioLPDistModule()
        mod._ml_available = True
        mel = mod._extract_log_mel(synthetic_wav)
        assert mel is not None
        dist = float(np.sqrt(np.mean((mel - mel) ** 2)))
        assert dist == 0.0


class TestAudioESTOI:
    def test_import(self):
        try:
            from ayase.modules.audio_estoi import AudioESTOIModule
            mod = AudioESTOIModule()
            assert mod.name == "audio_estoi"
        except ImportError:
            pytest.skip("pystoi not installed")


class TestAudioUTMOS:
    def test_import(self):
        from ayase.modules.audio_utmos import AudioUTMOSModule
        mod = AudioUTMOSModule()
        assert mod.name == "audio_utmos"



import codecs
import json
import os
from subprocess import call

import requests
from PIL import Image
from torchvision.datasets import VisionDataset

from .flores_langs import flores_languages

GITHUB_DATA_PATH = (
    "https://raw.githubusercontent.com/visheratin/nllb-clip/main/data/flickr30k-200/"
)
SUPPORTED_LANGUAGES = flores_languages

IMAGE_INDEX_FILENAME = "filenames.txt"

CAPTIONS_FILENAME_TEMPLATE = "{}.txt"
OUTPUT_FILENAME_TEMPLATE = "flickr30k_200-{}.json"

IMAGES_DOWNLOAD_URL = "https://nllb-data.com/test/flickr30k/images.tar.gz"


class Flickr30k_200(VisionDataset):
    def __init__(self, root, ann_file, transform=None, target_transform=None):
        super().__init__(root, transform=transform, target_transform=target_transform)
        self.ann_file = os.path.expanduser(ann_file)
        with codecs.open(ann_file, "r", encoding="utf-8") as fp:
            data = json.load(fp)
        self.data = [
            (img_path, txt)
            for img_path, txt in zip(data["image_paths"], data["annotations"])
        ]

    def __getitem__(self, index):
        img, captions = self.data[index]

        # Image
        img = Image.open(img).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)

        # Captions
        target = [
            captions,
        ]
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self) -> int:
        return len(self.data)


def _get_lines(url):
    response = requests.get(url, timeout=30)
    return response.text.splitlines()


def _download_images(out_path):
    os.makedirs(out_path, exist_ok=True)
    print("Downloading images")
    call(f"wget {IMAGES_DOWNLOAD_URL} -O images.tar.gz", shell=True)
    call(f"tar -xzf images.tar.gz -C {out_path}", shell=True)
    call("rm images.tar.gz", shell=True)


def create_annotation_file(root, lang_code):
    if lang_code not in SUPPORTED_LANGUAGES:
        raise ValueError(
            f"Language code {lang_code} not supported. Supported languages are {SUPPORTED_LANGUAGES}"
        )
    data_dir = os.path.join(root, "flickr30k-200")
    if not os.path.exists(data_dir):
        _download_images(data_dir)
    images_dir = os.path.join(root, "flickr30k-200", "images")
    print("Downloading flickr30k-200 index file")
    download_path = os.path.join(GITHUB_DATA_PATH, IMAGE_INDEX_FILENAME)
    target_images = _get_lines(download_path)

    print("Downloading flickr30k-200 captions:", lang_code)
    captions_path = GITHUB_DATA_PATH
    download_path = os.path.join(
        captions_path, CAPTIONS_FILENAME_TEMPLATE.format(lang_code)
    )
    target_captions = _get_lines(download_path)

    number_of_missing_images = 0
    valid_images, valid_annotations, valid_indicies = [], [], []
    for i, (img, txt) in enumerate(zip(target_images, target_captions)):
        image_path = os.path.join(images_dir, img)
        if not os.path.exists(image_path):
            print("Missing image file", img)
            number_of_missing_images += 1
            continue

        valid_images.append(image_path)
        valid_annotations.append(txt)
        valid_indicies.append(i)

    if number_of_missing_images > 0:
        print(f"*** WARNING *** missing {number_of_missing_images} files.")

    with codecs.open(
        os.path.join(root, OUTPUT_FILENAME_TEMPLATE.format(lang_code)),
        "w",
        encoding="utf-8",
    ) as fp:
        json.dump(
            {
                "image_paths": valid_images,
                "annotations": valid_annotations,
                "indicies": valid_indicies,
            },
            fp,
            ensure_ascii=False,
        )



"""Tests for knowledge_graph module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_knowledge_graph_basics():
    from ayase.modules.knowledge_graph import KnowledgeGraphModule
    _test_module_basics(KnowledgeGraphModule, "knowledge_graph")

def test_knowledge_graph_image(image_sample):
    from ayase.modules.knowledge_graph import KnowledgeGraphModule
    image_sample.quality_metrics = QualityMetrics()
    m = KnowledgeGraphModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_knowledge_graph_video(video_sample):
    from ayase.modules.knowledge_graph import KnowledgeGraphModule
    video_sample.quality_metrics = QualityMetrics()
    m = KnowledgeGraphModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
COCO dataset which returns image_id for evaluation.

Mostly copy-paste from https://github.com/pytorch/vision/blob/13b35ff/references/detection/coco_utils.py
"""
from pathlib import Path

import datasets.transforms as T

import torch
import torch.utils.data
from pycocotools import mask as coco_mask
from util.misc import get_local_rank, get_local_size

from .coco import CocoDetection, make_coco_transforms, make_coco_transforms_lsj
from .torchvision_datasets import CocoDetection as TvCocoDetection


def build(image_set, args):
    root = Path(args.coco_path)
    assert root.exists(), f"provided Objects365 path {root} does not exist"
    mode = "instances"
    PATHS = {
        "train": (
            root / "train",
            root / "annotations" / "zhiyuan_objv2_train_fixmiss.json",
        ),
        "val": (root / "val", root / "annotations" / "zhiyuan_objv2_val.json"),
    }

    img_folder, ann_file = PATHS[image_set]
    if args.lsj:
        coco_transform = make_coco_transforms_lsj(image_set, args.lsj_img_size)
    else:
        coco_transform = make_coco_transforms(image_set, args.bigger)
    dataset = CocoDetection(
        img_folder,
        ann_file,
        transforms=coco_transform,
        return_masks=args.masks,
        cache_mode=args.cache_mode,
        local_rank=get_local_rank(),
        local_size=get_local_size(),
    )
    return dataset



"""ChipQA module.

This module runs the ChipQA feature extractor and LIVE-Livestream SVR from the
bundled source tree.
"""

import logging
import re
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

_CHIPQA_BOOTSTRAP = r"""
import importlib
import runpy
import sys
from pathlib import Path

repo = Path(sys.argv[1])
input_folder = sys.argv[2]
results_folder = sys.argv[3]
sys.path.insert(0, str(repo.parent))
sys.modules["niqe"] = importlib.import_module("chipqa.niqe")
sys.modules["save_stats"] = importlib.import_module("chipqa.save_stats")
sys.argv = [str(repo / "chipqa.py"), input_folder, results_folder]
runpy.run_path(str(repo / "chipqa.py"), run_name="__main__")
"""


def _has_required_paths(base_dir: Path, required_paths: list[str]) -> bool:
    return all((base_dir / relative_path).exists() for relative_path in required_paths)


def _parse_last_float(text: str) -> float | None:
    matches = re.findall(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?", text)
    if not matches:
        return None
    try:
        return float(matches[-1])
    except ValueError:
        return None


def _python_executable(config_value: str | None) -> str:
    return str(config_value) if config_value else sys.executable


def _run_python_inline(
    python_bin: str,
    code: str,
    args: list[str],
    cwd: Path,
    timeout_sec: int,
) -> subprocess.CompletedProcess[str]:
    return subprocess.run(
        [python_bin, "-c", code, *args],
        cwd=str(cwd),
        capture_output=True,
        text=True,
        timeout=timeout_sec,
    )


def _third_party_path(*parts: str) -> Path:
    return Path(__file__).resolve().parents[1].joinpath("third_party", *parts)


class ChipQAModule(PipelineModule):
    name = "chipqa"
    description = "ChipQA no-reference video quality via official feature extractor and LIVE-Livestream SVR"
    default_config = {
        "repo_path": None,
        "python_executable": None,
        "timeout_sec": 1800,
        "warning_threshold": None,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.repo_path = self.config.get("repo_path")
        self.python_bin = _python_executable(self.config.get("python_executable"))
        self.timeout_sec = int(self.config.get("timeout_sec", 1800))
        self.warning_threshold = self.config.get("warning_threshold")
        self._backend = None
        self._repo_dir: Path | None = None

    def setup(self) -> None:
        try:
            import cv2  # noqa: F401
            import joblib  # noqa: F401
            import matplotlib  # noqa: F401
            import numba  # noqa: F401
            import numpy  # noqa: F401
            import scipy  # noqa: F401
            import sklearn  # noqa: F401
        except ImportError as exc:
            logger.warning("ChipQA unavailable: missing dependency: %s", exc)
            return

        repo_dir = (
            Path(self.repo_path).resolve()
            if self.repo_path
            else _third_party_path("hdr_chipqa", "chipqa")
        )
        required = [
            "chipqa.py",
            "testing.py",
            "LIVE_Livestream_trained_svr.z",
            "LIVE_Livestream_fitted_scaler.z",
            "frames_modelparameters.mat",
            "niqe.py",
            "save_stats.py",
        ]
        if not _has_required_paths(repo_dir, required):
            logger.warning("ChipQA unavailable: required source files are missing.")
            return

        self._repo_dir = repo_dir
        self._backend = "chipqa"

    def process(self, sample: Sample) -> Sample:
        if self._backend != "chipqa" or self._repo_dir is None:
            return sample
        if not sample.is_video or sample.path.suffix.lower() != ".mp4":
            return sample

        try:
            score = self._run_chipqa(sample.path)
            if score is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.chipqa_score = score

            if self.warning_threshold is not None and score < self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low ChipQA score: {score:.3f}",
                        details={"chipqa_score": score},
                    )
                )
        except Exception as exc:
            logger.warning("ChipQA failed for %s: %s", sample.path, exc)

        return sample

    def _run_chipqa(self, sample_path: Path) -> float | None:
        assert self._repo_dir is not None
        with tempfile.TemporaryDirectory() as tmp_dir_str:
            tmp_dir = Path(tmp_dir_str)
            input_dir = tmp_dir / "input"
            output_dir = tmp_dir / "output"
            input_dir.mkdir()
            output_dir.mkdir()

            staged_path = input_dir / sample_path.name
            shutil.copy2(sample_path, staged_path)

            extract = _run_python_inline(
                self.python_bin,
                _CHIPQA_BOOTSTRAP,
                [str(self._repo_dir), str(input_dir), str(output_dir)],
                cwd=self._repo_dir.parent,
                timeout_sec=self.timeout_sec,
            )
            if extract.returncode != 0:
                logger.warning("ChipQA feature extraction failed: %s", extract.stderr.strip())
                return None

            feature_path = output_dir / f"{sample_path.stem}.z"
            if not feature_path.exists():
                logger.warning("ChipQA did not produce feature file for %s", sample_path.name)
                return None

            predict = subprocess.run(
                [self.python_bin, "testing.py", str(feature_path)],
                cwd=str(self._repo_dir),
                capture_output=True,
                text=True,
                timeout=self.timeout_sec,
            )
            if predict.returncode != 0:
                logger.warning("ChipQA prediction failed: %s", predict.stderr.strip())
                return None
            return _parse_last_float(predict.stdout)
